#include <stdlib.h>
#include <new>

#include "log.h"
#include "memory.h"
#include "common.h"
#include "error.h"
#include "osd.h"


// ubN蓖
#define EMPTYROM	(RomB[0])	// EmptyRom
#define EMPTYRAM	(RomB[1])	// EmptyRam
#define EXTROM0		(RomB[2])	// ExtRom
#define EXTROM1		(RomB[3])	// ExtRom+0x2000
#define MAINROM0	(RomB[4])	// MainRom
#define MAINROM1	(RomB[5])	// MainRom+0x2000
#define MAINROM2	(RomB[6])	// MainRom+0x4000
#define MAINROM3	(RomB[7])	// MainRom+0x6000
#define MAINROM4	(RomB[8])	// MainRom+0x8000
#define MAINROM5	(RomB[9])	// MainRom+0xa000
#define MAINROM6	(RomB[10])	// MainRom+0xc000
#define MAINROM7	(RomB[11])	// MainRom+0xe000
#define CGROM1		(RomB[12])	// CGRom1
#define CGROM2		(RomB[13])	// CGRom2			CGRom1+0x2000
#define KANJIROM0	(RomB[14])	// KanjiRom			SysRom2+0x8000
#define KANJIROM1	(RomB[15])	// KanjiRom+0x2000	SysRom2+0xa000
#define KANJIROM2	(RomB[16])	// KanjiRom+0x4000	SysRom2+0xc000
#define KANJIROM3	(RomB[17])	// KanjiRom+0x6000	SysRom2+0xe000
#define VOICEROM0	(RomB[18])	// VoiceRom			SysRom2+0x4000
#define VOICEROM1	(RomB[19])	// VoiceRom+0x2000	ysRom2+0x6000
#define SRMENROM0	(RomB[20])	//					SysRom2
#define SRMENROM1	(RomB[21])	//					SysRom2+0x2000

#define MAINRAM0	(RamB[0])	// IntRam
#define MAINRAM1	(RamB[1])	// IntRam+0x2000
#define MAINRAM2	(RamB[2])	// IntRam+0x4000
#define MAINRAM3	(RamB[3])	// IntRam+0x6000
#define MAINRAM4	(RamB[4])	// IntRam+0x8000
#define MAINRAM5	(RamB[5])	// IntRam+0xa000
#define MAINRAM6	(RamB[6])	// IntRam+0xc000
#define MAINRAM7	(RamB[7])	// IntRam+0xe000
#define EXTRAM0		(RamB[8])	// ExtRam
#define EXTRAM1		(RamB[9])	// ExtRam+0x2000
#define EXTRAM2		(RamB[10])	// ExtRam+0x4000
#define EXTRAM3		(RamB[11])	// ExtRam+0x6000
#define EXTRAM4		(RamB[12])	// ExtRam+0x8000
#define EXTRAM5		(RamB[13])	// ExtRam+0xa000
#define EXTRAM6		(RamB[14])	// ExtRam+0xc000
#define EXTRAM7		(RamB[15])	// ExtRam+0xe000
#define SOLRAM		(RamB[11])	// EXTRAM3Ƌp


#define pEXTROM0	(UseExtRom ? &EXTROM0 : &EMPTYROM)
#define pEXTROM1	(UseExtRom ? &EXTROM1 : &EMPTYROM)

#define pKNJROM0	(kj_rom ? ( kj_LR ? &KANJIROM2 : &KANJIROM0 ) : &VOICEROM0 )
#define pKNJROM1	(kj_rom ? ( kj_LR ? &KANJIROM3 : &KANJIROM1 ) : &VOICEROM1 )


// Rg[WX^l
#define	INIT_RF0	(0x71)
#define INIT_RF1	(0xdd)
#define INIT_RF2	(0x50)





////////////////////////////////////////////////////////////////
// ROM
////////////////////////////////////////////////////////////////
// PC-6001                                  ROMt@C  CRC32
const MEM6::ROMINFO MEM60::AMAINROM[] = { { "BASICROM.60", 0x54c03109 } };
const MEM6::ROMINFO MEM60::ACGROM1[]  = { { "CGROM60.60",  0xb0142d32 } };

// PC-6001A                                 ROMt@C  CRC32
const MEM6::ROMINFO MEM61::AMAINROM[] = { { "BASICROM.61", 0xfa8e88d9 } };
const MEM6::ROMINFO MEM61::ACGROM1[]  = { { "CGROM60.61",  0x49c21d08 } };

// PC-6001mk2                               ROMt@C  CRC32
const MEM6::ROMINFO MEM62::AMAINROM[] = { { "BASICROM.62", 0x950ac401 },	// O
										  { "BASICROM.62", 0xd7e61957 } };	// 
const MEM6::ROMINFO MEM62::ACGROM1[]  = { { "CGROM60.62",  0x81eb5d95 } };
const MEM6::ROMINFO MEM62::ACGROM2[]  = { { "CGROM60m.62", 0x3ce48c33 } };
const MEM6::ROMINFO MEM62::AKANJI[]   = { { "KANJIROM.62", 0x20c8f3eb } };
const MEM6::ROMINFO MEM62::AVOICE[]   = { { "VOICEROM.62", 0x49b4f917 } };

// PC-6601                                  ROMt@C  CRC32
const MEM6::ROMINFO MEM66::AMAINROM[] = { { "BASICROM.66", 0xc0b01772 } };
const MEM6::ROMINFO MEM66::ACGROM1[]  = { { "CGROM60.66",  0xd2434f29 } };
const MEM6::ROMINFO MEM66::ACGROM2[]  = { { "CGROM66.66",  0x3ce48c33 } };
const MEM6::ROMINFO MEM66::AKANJI[]   = { { "KANJIROM.66", 0x20c8f3eb } };
const MEM6::ROMINFO MEM66::AVOICE[]   = { { "VOICEROM.66", 0x91d078c1 } };

// PC-6001mk2SR                             ROMt@C  CRC32
// PC-6601SR                                ROMt@C  CRC32


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
//                                  ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM6::IEMPTROM  = { NULL,     0,                 0x02000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM6::IEMPTRAM  = { NULL,     0,                 0x02000, 0xff,  0, 0 };
const MEM6::MEMINFO MEM6::IEXTROM   = { NULL,     0,                 0x20000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM6::IEXTRAM   = { NULL,     0,                 0x20000, 0x00,  0, 0 };
										// MEGA-ROM/RAMJ[gbWΉ̂
										// gROM/RAM̈128KBmۂĂ

// ----------------------------------------------------------------------------------------
// m̃J[gbW                   ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM6::ISOLRAM   = { NULL,     0,                 0x02000, 0xff,  1, 1 };
// ----------------------------------------------------------------------------------------

// PC-6001                              ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM60::IMAINROM = { AMAINROM, COUNTOF(AMAINROM), 0x04000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM60::ICGROM1  = { ACGROM1,  COUNTOF(ACGROM1),  0x01000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM60::IINTRAM  = { NULL,     0,                 0x04000, 0x00,  0, 0 };

// PC-6001A                             ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM61::IMAINROM = { AMAINROM, COUNTOF(AMAINROM), 0x04000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM61::ICGROM1  = { ACGROM1,  COUNTOF(ACGROM1),  0x01000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM61::IINTRAM  = { NULL,     0,                 0x04000, 0x00,  0, 0 };

// PC-6001mk2                           ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM62::IMAINROM = { AMAINROM, COUNTOF(AMAINROM), 0x08000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::ICGROM1  = { ACGROM1,  COUNTOF(ACGROM1),  0x02000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::ICGROM2  = { ACGROM2,  COUNTOF(ACGROM2),  0x02000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IKANJI   = { AKANJI,   COUNTOF(AKANJI),   0x08000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IVOICE   = { AVOICE,   COUNTOF(AVOICE),   0x04000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM62::IINTRAM  = { NULL,     0,                 0x10000, 0x00,  0, 0 };

// PC-6601                              ROM   ROMvf      TCY   l Wait(Read/Write)
const MEM6::MEMINFO MEM66::IMAINROM = { AMAINROM, COUNTOF(AMAINROM), 0x08000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::ICGROM1  = { ACGROM1,  COUNTOF(ACGROM1),  0x02000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::ICGROM2  = { ACGROM2,  COUNTOF(ACGROM2),  0x02000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IKANJI   = { AKANJI,   COUNTOF(AKANJI),   0x08000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IVOICE   = { AVOICE,   COUNTOF(AVOICE),   0x04000, 0xff,  1, 1 };
const MEM6::MEMINFO MEM66::IINTRAM  = { NULL,     0,                 0x10000, 0x00,  0, 0 };

// PC-6001mk2SR                         ROM   ROMvf      TCY   l Wait(Read/Write)
// PC-6601SR                            ROM   ROMvf      TCY   l Wait(Read/Write)


//--------------------------------------------------------------
// ubN
//--------------------------------------------------------------

////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
MemBlock::MemBlock( void ) : MB(NULL), RWait(0), WWait(0), WPt(false)
{
	INITARRAY( Name, '\0' );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
MemBlock::~MemBlock( void ){}


////////////////////////////////////////////////////////////////
// ubNݒ
//
// :	name	ubN
//			mem		ubNւ̃|C^
//			rw		ANZXEFCg(Ǎ)
//			ww		ANZXEFCg()
//			prt		CgveNgtO true:Zbg falseF
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetMemory( const char *name, BYTE *mem, int rw, int ww, bool prt )
{
	strncpy( Name, name, 32 );
	MB    = mem;
	RWait = rw;
	WWait = ww;
	WPt = prt;
}


////////////////////////////////////////////////////////////////
// ANZXEFCgݒ
//
// :	rw		ANZXEFCg(Ǎ)
//			ww		ANZXEFCg()
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetWait( int rw, int ww )
{
	RWait = rw;
	WWait = ww;
}


////////////////////////////////////////////////////////////////
// ANZXEFCg擾
//
// :	Ȃ
// Ԓl:	bit0-7	ǍݎEFCg
//			bit8-15	ݎEFCg
////////////////////////////////////////////////////////////////
int MemBlock::GetWait( void )
{
	return ((WWait&0xff)<<8) | (RWait&0xff);
}


////////////////////////////////////////////////////////////////
// CgveNgݒ
//
// :	prt		CgveNgtO true:Zbg falseF
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void MemBlock::SetProtect( bool prt )
{
	WPt = prt;
}


////////////////////////////////////////////////////////////////
// CgveNg擾
//
// :	Ȃ
// Ԓl:	bool	true:Zbg falseF
////////////////////////////////////////////////////////////////
bool MemBlock::GetProtect( void )
{
	return WPt;
}


////////////////////////////////////////////////////////////////
// ubN擾
//
// :	Ȃ
// Ԓl:	char *	ubNւ̃|C^
////////////////////////////////////////////////////////////////
const char *MemBlock::GetName( void )
{
	return Name;
}


////////////////////////////////////////////////////////////////
// [h
////////////////////////////////////////////////////////////////
BYTE MemBlock::Read( WORD addr, int *wcnt )
{
	if( wcnt ) (*wcnt) += RWait;
	return MB[addr&0x1fff];
}


////////////////////////////////////////////////////////////////
// Cg
////////////////////////////////////////////////////////////////
void MemBlock::Write( WORD addr, BYTE data, int *wcnt )
{
	if( wcnt ) (*wcnt) += WWait;
	if( !WPt ) MB[addr&0x1fff] = data;
}




////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
MEM6::MEM6( VM6 *vm, const ID& id ) : Device(vm,id),
	CGBank(false), UseExtRom(false), UseExtRam(false),
	MainRom(NULL), ExtRom(NULL), CGRom1(NULL), CGRom2(NULL),
	KanjiRom(NULL), VoiceRom(NULL), IntRam(NULL), ExtRam(NULL),
	EmptyRom(NULL), EmptyRam(NULL),
	M1Wait(1), EnableChkCRC(true),
	UseSol(false), SolBank(0)
{
	INITARRAY( Rm_blk, NULL );
	INITARRAY( Wm_blk, NULL );
	INITARRAY( FilePath, '\0' );
	
	MemTable.EmptRom = &MEM6::IEMPTROM;
	MemTable.EmptRam = &MEM6::IEMPTRAM;
	MemTable.ExtRom  = &MEM6::IEXTROM;
	MemTable.ExtRam  = &MEM6::IEXTRAM;
	MemTable.SolRam  = &MEM6::ISOLRAM;
}

MEM60::MEM60( VM6 *vm, const ID& id ) : MEM6(vm,id)
{
	RomTable.System1 = MEM60::AMAINROM;
	RomTable.CGRom1  = MEM60::ACGROM1;
	
	MemTable.IntRam  = &MEM60::IINTRAM;
	MemTable.System1 = &MEM60::IMAINROM;
	MemTable.CGRom1  = &MEM60::ICGROM1;
}

MEM61::MEM61( VM6 *vm, const ID& id ) : MEM60(vm,id)
{
	RomTable.System1 = MEM61::AMAINROM;
	RomTable.CGRom1  = MEM61::ACGROM1;
	
	MemTable.IntRam  = &MEM61::IINTRAM;
	MemTable.System1 = &MEM61::IMAINROM;
	MemTable.CGRom1  = &MEM61::ICGROM1;
}

MEM62::MEM62( VM6 *vm, const ID& id ) : MEM6(vm,id),
	cgrom(true), kj_rom(true), kj_LR(true), cgenable(true), cgaddr(3),
	Rf0(INIT_RF0), Rf1(INIT_RF1), Rf2(INIT_RF2)
{
	RomTable.System1 = MEM62::AMAINROM;
	RomTable.CGRom1  = MEM62::ACGROM1;
	RomTable.CGRom2  = MEM62::ACGROM2;
	RomTable.Kanji   = MEM62::AKANJI;
	RomTable.Voice   = MEM62::AVOICE;
	
	MemTable.IntRam  = &MEM62::IINTRAM;
	MemTable.System1 = &MEM62::IMAINROM;
	MemTable.CGRom1  = &MEM62::ICGROM1;
	MemTable.CGRom2  = &MEM62::ICGROM2;
	MemTable.Kanji   = &MEM62::IKANJI;
	MemTable.Voice   = &MEM62::IVOICE;
}

MEM66::MEM66( VM6 *vm, const ID& id ) : MEM62(vm,id)
{
	RomTable.System1 = MEM66::AMAINROM;
	RomTable.CGRom1  = MEM66::ACGROM1;
	RomTable.CGRom2  = MEM66::ACGROM2;
	RomTable.Kanji   = MEM66::AKANJI;
	RomTable.Voice   = MEM66::AVOICE;
	
	MemTable.IntRam  = &MEM66::IINTRAM;
	MemTable.System1 = &MEM66::IMAINROM;
	MemTable.CGRom1  = &MEM66::ICGROM1;
	MemTable.CGRom2  = &MEM66::ICGROM2;
	MemTable.Kanji   = &MEM66::IKANJI;
	MemTable.Voice   = &MEM66::IVOICE;
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
MEM6::~MEM6( void )
{
	if( MainRom )	delete[] MainRom;
	if( IntRam )	delete[] IntRam;
	if( CGRom1 )	delete[] CGRom1;
	if( ExtRom )	delete[] ExtRom;
	if( EmptyRom )	delete[] EmptyRom;
	if( EmptyRam )	delete[] EmptyRam;
}

MEM60::~MEM60( void ){}

MEM61::~MEM61( void ){}

MEM62::~MEM62( void )
{
	if( CGRom2 )	delete[] CGRom2;
	if( KanjiRom )	delete[] KanjiRom;
	if( VoiceRom )	delete[] VoiceRom;
}

MEM66::~MEM66( void ){}


////////////////////////////////////////////////////////////////
// gROM }Eg
////////////////////////////////////////////////////////////////
bool MEM6::MountExtRom( const char *filename )
{
	PRINTD( MEM_LOG, "[MEM][MountExtRom] -> %s -> ", filename );
	
	// }Egς݂ȂUJ
	if( UseExtRom ) UnmountExtRom();
	
	try{
		FILE *fp = FOPENEN( filename, "rb" );
		if( !fp ) throw Error::ExtRomMountFailed;
		fread( ExtRom, sizeof(BYTE), IEXTROM.Size, fp );
		fclose( fp );
		
		// t@CpXۑ
		strncpy( FilePath, filename, PATH_MAX );
	}
	// O
	catch( Error::Errno i ){
		Error::SetError( i );
		
		PRINTD( MEM_LOG, "Error\n" );
		
		UnmountExtRom();
		return false;
	}
	
	UseExtRom = true;
	
	PRINTD( MEM_LOG, "OK %s\n", FilePath );
	
	return true;
}


////////////////////////////////////////////////////////////////
// gROM A}Eg
////////////////////////////////////////////////////////////////
void MEM6::UnmountExtRom( void )
{
	PRINTD( MEM_LOG, "[MEM][UnmountExtRom]\n" );
	
	memset( ExtRom, IEXTROM.Init, IEXTROM.Size );
	*FilePath = '\0';
	
	UseExtRom = false;
}


////////////////////////////////////////////////////////////////
// gROMt@CpX擾
////////////////////////////////////////////////////////////////
const char *MEM6::GetFile( void )
{
	return FilePath;
}


////////////////////////////////////////////////////////////////
// mۂROMt@CǍ
////////////////////////////////////////////////////////////////
bool MEM6::AllocMemory( BYTE **buf, const MEMINFO *info, const char *path )
{
	PRINTD( MEM_LOG, "[MEM][AllocMemory] " );
	
	DWORD LoadSize = 0;
	int i = 0;
	bool ErrSize = false;
	bool ErrCrc = false;
	
	try{
		// m(ŏTCY8KB)
		if( *buf) delete [] *buf;
		*buf = new BYTE[max( info->Size, 0x2000 )];
		memset( *buf, info->Init, info->Size );
		
		// ROMȂ ȂRAM܂NULL
		if( !info->Rinf ) return true;
		
		do{
			PRINTD( MEM_LOG, "-> %s ", info->Rinf[i].FileName );
			
			// t@CǍ
			char fpath[PATH_MAX] = "";
			if( path ){
				strncpy( fpath, path, PATH_MAX );
				AddDelimiter( fpath );
			}
			strncat( fpath, info->Rinf[i].FileName, PATH_MAX );
			
			FILE *fp = FOPENEN( fpath, "rb" );
			if( fp ){
				LoadSize = fread( *buf, sizeof(BYTE), info->Size, fp );
				fclose( fp );
				
				// t@CTCY`FbN
				if( LoadSize != info->Size )
					ErrSize = true;
				// CRC`FbN
				// EnableChkCRC=false ܂ CRC=0̎̓`FbNȂ
				else if( EnableChkCRC && (info->Rinf[i].Crc != 0) &&
					( CalcCrc32( *buf, info->Size ) != info->Rinf[i].Crc ) )
					ErrCrc = true;
				else{
					PRINTD( MEM_LOG, "-> OK\n" );
					return true;
				}
			}
		}while( ++i < info->Rnum );
		
		if     ( ErrCrc )  throw Error::RomCrcNG;
		else if( ErrSize ) throw Error::RomSizeNG;
		else               throw Error::NoRom;
	}
	// new Ɏsꍇ
	catch( std::bad_alloc ){
		PRINTD( MEM_LOG, "-> MemAlloc Failed\n" );
		Error::SetError( Error::MemAllocFailed );
		return false;
	}
	// O
	catch( Error::Errno i ){
		PRINTD( MEM_LOG, "-> Failed\n" );
		
		Error::SetError( i );
		
		switch(i){
		case Error::NoRom:		// t@C̃I[vɎsꍇ
		case Error::RomSizeNG:	// TCYȂꍇ
		case Error::RomCrcNG:	// CRCȂꍇ
		default:				// J
			if( *buf) delete [] *buf;
			*buf = NULL;
		}
		return false;
	}
	
	return false;
}


////////////////////////////////////////////////////////////////
// SmۂROMt@CǍ
////////////////////////////////////////////////////////////////
bool MEM6::AllocAllMemory( const char *path, bool chkcrc )
{
	PRINTD( MEM_LOG, "[MEM][AllocAllMemory]\n" );
	
	EnableChkCRC = chkcrc;		// CRC`FbNL
	
	// 
	if( !AllocMemory( &EmptyRom, MemTable.EmptRom, NULL ) ) return false;
	if( !AllocMemory( &EmptyRam, MemTable.EmptRam, NULL ) ) return false;
	
	if( !AllocMemory( &ExtRom,   MemTable.ExtRom,  NULL ) ) return false;
	if( !AllocMemory( &ExtRam,   MemTable.ExtRam,  NULL ) ) return false;
	if( !AllocMemory( &IntRam,   MemTable.IntRam,  NULL ) ) return false;
	
	if( !AllocMemory( &MainRom,  MemTable.System1, path ) ) return false;
	if( !AllocMemory( &CGRom1,   MemTable.CGRom1,  path ) ) return false;
	
	// 62,66
	if( MemTable.CGRom2 ) if( !AllocMemory( &CGRom2,   MemTable.CGRom2, path ) ) return false;
	if( MemTable.Kanji )  if( !AllocMemory( &KanjiRom, MemTable.Kanji,  path ) ) return false;
	if( MemTable.Voice )  if( !AllocMemory( &VoiceRom, MemTable.Voice,  path ) ) return false;
	
	SetRamValue();
	
	return true;
}


////////////////////////////////////////////////////////////////
// RAM̏lݒ
////////////////////////////////////////////////////////////////
void MEM60::SetRamValue( void )
{
	int addr = 0;
	for( int i=0; i < (int)IINTRAM.Size/128; i++ ){
		for( int j=0; j < 64; j++ ) IntRam[addr++] = 0;
		for( int j=0; j < 64; j++ ) IntRam[addr++] = 0xff;
	}
}

void MEM62::SetRamValue( void )
{
	int addr = 0;
	for( int i=0; i < (int)IINTRAM.Size/256; i++ ){
		for( int j=0; j < 64; j++ ){
			IntRam[addr++] = 0;
			IntRam[addr++] = 0xff;
		}
		for( int j=0; j < 64; j++ ){
			IntRam[addr++] = 0xff;
			IntRam[addr++] = 0;
		}
	}
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
bool MEM6::Init( bool useexram, bool usesol )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	UseExtRam = useexram;	// ORAML
	UseSol    = usesol;		// m̃J[gbWL
	
	// ubNݒ
	// Ƃ肠SEmptyɐݒ
	for( int i=0; i<MAXRMB; i++ )
		RomB[i].SetMemory( "EmptyRom", EmptyRom, MemTable.EmptRom->WaitR, MemTable.EmptRom->WaitW, true );
	for( int i=0; i<MAXWMB; i++ )
		RamB[i].SetMemory( "EmptyRam", EmptyRam, MemTable.EmptRam->WaitR, MemTable.EmptRam->WaitW, true );
	
	// gROM̈
	EXTROM0.SetMemory( "ExtRom0", ExtRom,        MemTable.ExtRom->WaitR, MemTable.ExtRom->WaitW, true );
	EXTROM1.SetMemory( "ExtRom1", ExtRom+0x2000, MemTable.ExtRom->WaitR, MemTable.ExtRom->WaitW, true );
	
	// ORAMݒ(r)
	if( UseSol ){			// m̃J[gbW
		EXTRAM0.SetMemory( "ExtRam0", ExtRam,        MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM1.SetMemory( "ExtRam1", ExtRam+0x2000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		SOLRAM.SetMemory ( "SolRam",  ExtRam+0x6000, MemTable.SolRam->WaitR, MemTable.SolRam->WaitW, false );
	}else if( UseExtRam ){	// RAMJ[gbW
		EXTRAM0.SetMemory( "ExtRam0", ExtRam,        MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM1.SetMemory( "ExtRam1", ExtRam+0x2000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM2.SetMemory( "ExtRam2", ExtRam+0x4000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM3.SetMemory( "ExtRam3", ExtRam+0x6000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM4.SetMemory( "ExtRam4", ExtRam+0x8000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM5.SetMemory( "ExtRam5", ExtRam+0xa000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM6.SetMemory( "ExtRam6", ExtRam+0xc000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
		EXTRAM7.SetMemory( "ExtRam7", ExtRam+0xe000, MemTable.ExtRam->WaitR, MemTable.ExtRam->WaitW, false );
	}
	
	 // @ʏ
	if( !InitSpecific() ) return false;
	
	return true;
}


////////////////////////////////////////////////////////////////
// @ʏ
////////////////////////////////////////////////////////////////
bool MEM60::InitSpecific( void )
{
	PRINTD( MEM_LOG, "[MEM][InitSpecific]\n" );
	
	// ubNݒ
	// BASIC ROM
	MAINROM0.SetMemory( "MainRom0", MainRom,        MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	MAINROM1.SetMemory( "MainRom1", MainRom+0x2000, MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	
	// CG ROM
	CGROM1.SetMemory  ( "CGRom1",   CGRom1,         MemTable.CGRom1->WaitR,  MemTable.CGRom1->WaitW,  true );
	
	// RAM
	MAINRAM0.SetMemory( "IntRam0",  IntRam,         MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM1.SetMemory( "IntRam1",  IntRam+0x2000,  MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	
	return true;
}

bool MEM62::InitSpecific( void )
{
	PRINTD( MEM_LOG, "[MEM][InitSpecific]\n" );
	
	// ubNݒ
	// BASIC ROM
	MAINROM0.SetMemory ( "MainRom0", MainRom,         MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	MAINROM1.SetMemory ( "MainRom1", MainRom+0x2000,  MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	MAINROM2.SetMemory ( "MainRom2", MainRom+0x4000,  MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	MAINROM3.SetMemory ( "MainRom3", MainRom+0x6000,  MemTable.System1->WaitR, MemTable.System1->WaitW, true );
	
	// CG ROM
	CGROM1.SetMemory   ( "CGRom1",   CGRom1,          MemTable.CGRom1->WaitR,  MemTable.CGRom1->WaitW,  true );
	CGROM2.SetMemory   ( "CGRom2",   CGRom2,          MemTable.CGRom2->WaitR,  MemTable.CGRom2->WaitW,  true );
	
	// ROM
	KANJIROM0.SetMemory( "KanjRom0", KanjiRom,        MemTable.Kanji->WaitR,   MemTable.Kanji->WaitW,   true );
	KANJIROM1.SetMemory( "KanjRom1", KanjiRom+0x2000, MemTable.Kanji->WaitR,   MemTable.Kanji->WaitW,   true );
	KANJIROM2.SetMemory( "KanjRom2", KanjiRom+0x4000, MemTable.Kanji->WaitR,   MemTable.Kanji->WaitW,   true );
	KANJIROM3.SetMemory( "KanjRom3", KanjiRom+0x6000, MemTable.Kanji->WaitR,   MemTable.Kanji->WaitW,   true );
	
	// ROM
	VOICEROM0.SetMemory( "VoicRom0", VoiceRom,        MemTable.Voice->WaitR,   MemTable.Voice->WaitW,   true );
	VOICEROM1.SetMemory( "VoicRom1", VoiceRom+0x2000, MemTable.Voice->WaitR,   MemTable.Voice->WaitW,   true );
	
	// RAM
	MAINRAM0.SetMemory ( "IntRam0",  IntRam,          MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM1.SetMemory ( "IntRam1",  IntRam+0x2000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM2.SetMemory ( "IntRam2",  IntRam+0x4000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM3.SetMemory ( "IntRam3",  IntRam+0x6000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM4.SetMemory ( "IntRam4",  IntRam+0x8000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM5.SetMemory ( "IntRam5",  IntRam+0xa000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM6.SetMemory ( "IntRam6",  IntRam+0xc000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	MAINRAM7.SetMemory ( "IntRam7",  IntRam+0xe000,   MemTable.IntRam->WaitR,  MemTable.IntRam->WaitW,  false );
	
	return true;
}


////////////////////////////////////////////////////////////////
// pb`
////////////////////////////////////////////////////////////////
void MEM6::Patch( void )
{
	// 60,61,62,66 
	
	// WCXeBbN̓[` oOC(sS)
	MainRom[0x1cb4] = 0xe6;	// AND A,x
	MainRom[0x1cb5] = 0xbf;
}


////////////////////////////////////////////////////////////////
// Zbg
////////////////////////////////////////////////////////////////
void MEM6::Reset(){}

void MEM60::Reset( void )
{
	PRINTD( MEM_LOG, "[MEM][Reset]\n" );
	
	CGBank = false;	// CG ROM BANK 
	
	// CGROM̌㔼4KBɊgROM̓eRs[
	// (oN؂ւ8KBPʂōs)
	// gROMꍇCGROM̃C[WƂȂ炵
	memcpy( &CGRom1[0x1000], UseExtRom ? &ExtRom[0x3000] : &CGRom1[0x0000], 0x1000 );
	
	SetMemBlock();
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( 0 );	// oN
	// -------------------------------------------------------------
}

void MEM62::Reset( void )
{
	PRINTD( MEM_LOG, "[MEM][Reset]\n" );
	
	CGBank = false;	// CG ROM BANK 
	Rf0 = INIT_RF0;	// Rg[WX^lݒ
	Rf1 = INIT_RF1;	// Rg[WX^lݒ
	Rf2 = INIT_RF2;	// Rg[WX^lݒ
	
	SetMemBlockR( Rf0, Rf1 );
	SetMemBlockW( Rf2 );
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( 0 );	// oN
	// -------------------------------------------------------------
}


////////////////////////////////////////////////////////////////
// tFb`(M1)
////////////////////////////////////////////////////////////////
BYTE MEM6::Fetch( WORD addr, int *m1wait )
{
	PRINTD( MEM_LOG, "[MEM][Fetch] -> %04X", addr );
	
	// M1EFCgǉ
	if( m1wait ) (*m1wait) += M1Wait;
	
	return Rm_blk[addr>>13]->Read( addr );
}


////////////////////////////////////////////////////////////////
// [h
////////////////////////////////////////////////////////////////
BYTE MEM6::Read( WORD addr, int *wcnt )
{
	PRINTD( MEM_LOG, "[MEM][Read] -> %04X", addr );
	
	return Rm_blk[addr>>13]->Read( addr, wcnt );
}


////////////////////////////////////////////////////////////////
// Cg
////////////////////////////////////////////////////////////////
void MEM6::Write( WORD addr, BYTE data, int *wcnt )
{
	PRINTD( MEM_LOG, "[MEM][Write] %04X:%02X -> %p[%04X]'%c'", addr, data, Wm_blk[addr>>13], addr&0x1fff, data );
	
	Wm_blk[addr>>13]->Write( addr, data, wcnt );
	
	// /ORAMƂ݂̏ꍇ͂ЂƂ܂
}


////////////////////////////////////////////////////////////////
// [h̃ubNw(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetMemBlockR( BYTE mem1, BYTE mem2 )
{
	PRINTD( MEM_LOG, "[MEM][SetMemBlockR]\n" );
	
	// Port F0H
	switch( mem1 & 0x0f ){	// RF0 (0x0000 - 0x3fff)
		case 0x00:	Rm_blk[0] = &EMPTYROM;	Rm_blk[1] = &EMPTYROM;	break;
		case 0x01:	Rm_blk[0] = &MAINROM0;	Rm_blk[1] = &MAINROM1;	break;
		case 0x02:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x03:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = pEXTROM1;	break;
		case 0x04:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pEXTROM0;	break;
		case 0x05:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = &MAINROM1;	break;
		case 0x06:	Rm_blk[0] = &MAINROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x07:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pEXTROM1;	break;
		case 0x08:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = pEXTROM0;	break;
		case 0x09:	Rm_blk[0] = pEXTROM1;	Rm_blk[1] = &MAINROM1;	break;
		case 0x0a:	Rm_blk[0] = &MAINROM0;	Rm_blk[1] = pEXTROM1;	break;
		case 0x0b:	Rm_blk[0] = pEXTROM0;	Rm_blk[1] = pKNJROM1;	break;
		case 0x0c:	Rm_blk[0] = pKNJROM0;	Rm_blk[1] = pEXTROM0;	break;
		case 0x0d:	Rm_blk[0] = &MAINRAM0;	Rm_blk[1] = &MAINRAM1;	break;
		case 0x0e:	Rm_blk[0] = &EXTRAM0;	Rm_blk[1] = &EXTRAM1;	break;
		case 0x0f:	Rm_blk[0] = &EMPTYROM;	Rm_blk[1] = &EMPTYROM;	break;
	}
	switch( mem1 & 0xf0 ){	// RF0 (0x4000 - 0x7fff)
		case 0x00:	Rm_blk[2] = &EMPTYROM;	Rm_blk[3] = &EMPTYROM;	break;
		case 0x10:	Rm_blk[2] = &MAINROM2;	Rm_blk[3] = &MAINROM3;	break;
		case 0x20:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = pKNJROM1;	break;
		case 0x30:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = UseSol ? &SOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0x40:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = pEXTROM0;	break;
		case 0x50:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = &MAINROM3;	break;
		case 0x60:	Rm_blk[2] = &MAINROM2;	Rm_blk[3] = pKNJROM1;	break;
		case 0x70:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = UseSol ? &SOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0x80:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = pEXTROM0;	break;
		case 0x90:	Rm_blk[2] = pEXTROM1;	Rm_blk[3] = &MAINROM3;	break;
		case 0xa0:	Rm_blk[2] = &MAINROM2;	Rm_blk[3] = UseSol ? &SOLRAM : pEXTROM1;	break;	// m̃J[gbWΉ
		case 0xb0:	Rm_blk[2] = pEXTROM0;	Rm_blk[3] = pKNJROM1;	break;
		case 0xc0:	Rm_blk[2] = pKNJROM0;	Rm_blk[3] = pEXTROM0;	break;
		case 0xd0:	Rm_blk[2] = &MAINRAM2;	Rm_blk[3] = &MAINRAM3;	break;
		case 0xe0:	Rm_blk[2] = &EXTRAM2;	Rm_blk[3] = &EXTRAM3;	break;
		case 0xf0:	Rm_blk[2] = &EMPTYROM;	Rm_blk[3] = &EMPTYROM;	break;
	}
	
	// Port F1H
	switch( mem2 & 0x0f ){	// RF1 (0x8000 - 0xbfff)
		case 0x00:	Rm_blk[4] = &EMPTYROM;	Rm_blk[5] = &EMPTYROM;	break;
		case 0x01:	Rm_blk[4] = &MAINROM0;	Rm_blk[5] = &MAINROM1;	break;
		case 0x02:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x03:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = pEXTROM1;	break;
		case 0x04:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pEXTROM0;	break;
		case 0x05:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = &MAINROM1;	break;
		case 0x06:	Rm_blk[4] = &MAINROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x07:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pEXTROM1;	break;
		case 0x08:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = pEXTROM0;	break;
		case 0x09:	Rm_blk[4] = pEXTROM1;	Rm_blk[5] = &MAINROM1;	break;
		case 0x0a:	Rm_blk[4] = &MAINROM0;	Rm_blk[5] = pEXTROM1;	break;
		case 0x0b:	Rm_blk[4] = pEXTROM0;	Rm_blk[5] = pKNJROM1;	break;
		case 0x0c:	Rm_blk[4] = pKNJROM0;	Rm_blk[5] = pEXTROM0;	break;
		case 0x0d:	Rm_blk[4] = &MAINRAM4;	Rm_blk[5] = &MAINRAM5;	break;
		case 0x0e:	Rm_blk[4] = &EXTRAM4;	Rm_blk[5] = &EXTRAM5;	break;
		case 0x0f:	Rm_blk[4] = &EMPTYROM;	Rm_blk[5] = &EMPTYROM;	break;
	}
	switch( mem2 & 0xf0 ){	// RF1 (0xc000 - 0xffff)
		case 0x00:	Rm_blk[6] = &EMPTYROM;	Rm_blk[7] = &EMPTYROM;	break;
		case 0x10:	Rm_blk[6] = &MAINROM2;	Rm_blk[7] = &MAINROM3;	break;
		case 0x20:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = pKNJROM1;	break;
		case 0x30:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = pEXTROM1;	break;
		case 0x40:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pEXTROM0;	break;
		case 0x50:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = &MAINROM3;	break;
		case 0x60:	Rm_blk[6] = &MAINROM2;	Rm_blk[7] = pKNJROM1;	break;
		case 0x70:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pEXTROM1;	break;
		case 0x80:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = pEXTROM0;	break;
		case 0x90:	Rm_blk[6] = pEXTROM1;	Rm_blk[7] = &MAINROM3;	break;
		case 0xa0:	Rm_blk[6] = &MAINROM2;	Rm_blk[7] = pEXTROM1;	break;
		case 0xb0:	Rm_blk[6] = pEXTROM0;	Rm_blk[7] = pKNJROM1;	break;
		case 0xc0:	Rm_blk[6] = pKNJROM0;	Rm_blk[7] = pEXTROM0;	break;
		case 0xd0:	Rm_blk[6] = &MAINRAM6;	Rm_blk[7] = &MAINRAM7;	break;
		case 0xe0:	Rm_blk[6] = &EXTRAM6;	Rm_blk[7] = &EXTRAM7;	break;
		case 0xf0:	Rm_blk[6] = &EMPTYROM;	Rm_blk[7] = &EMPTYROM;	break;
	}
	
	if( CGBank ) Rm_blk[cgaddr] = cgrom ? &CGROM1 : &CGROM2;
	
	// WX^ۑ
	Rf0 = mem1;
	Rf1 = mem2;
	
	#if (MEM_LOG)
	fprintf( stdout, "[MEM][SetMemBlockR] %02X %02X\n", mem1, mem2 );
	for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetReadMemBlk(i) );
	#endif
}



////////////////////////////////////////////////////////////////
// Cg̃ubNw(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetMemBlockW( BYTE data )
{
	PRINTD( MEM_LOG, "[MEM][SetMemBlockW] -> %02X\n", data );
	
	switch( data & 3 ){
	case 0: Wm_blk[0] = &EMPTYRAM;	Wm_blk[1] = &EMPTYRAM;	break;
	case 1: Wm_blk[0] = &MAINRAM0;	Wm_blk[1] = &MAINRAM1;	break;
	case 2: Wm_blk[0] = &EXTRAM0;	Wm_blk[1] = &EXTRAM1;	break;
	case 3: Wm_blk[0] = &MAINRAM0;	Wm_blk[1] = &MAINRAM1;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( (data>>2) & 3 ){
	case 0: Wm_blk[2] = &EMPTYRAM;	Wm_blk[3] = &EMPTYRAM;	break;
	case 1: Wm_blk[2] = &MAINRAM2;	Wm_blk[3] = UseSol ? &SOLRAM : &MAINRAM3;	break;	// m̃J[gbWΉ
	case 2: Wm_blk[2] = &EXTRAM2;	Wm_blk[3] = UseSol ? &SOLRAM : &EXTRAM3;	break;	// m̃J[gbWΉ
	case 3: Wm_blk[2] = &MAINRAM2;	Wm_blk[3] = UseSol ? &SOLRAM : &MAINRAM3;	break;	// m̃J[gbWΉ
																						// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( (data>>4) & 3 ){
	case 0: Wm_blk[4] = &EMPTYRAM;	Wm_blk[5] = &EMPTYRAM;	break;
	case 1: Wm_blk[4] = &MAINRAM4;	Wm_blk[5] = &MAINRAM5;	break;
	case 2: Wm_blk[4] = &EXTRAM4;	Wm_blk[5] = &EXTRAM5;	break;
	case 3: Wm_blk[4] = &MAINRAM4;	Wm_blk[5] = &MAINRAM5;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	switch( (data>>6) & 3 ){
	case 0: Wm_blk[6] = &EMPTYRAM;	Wm_blk[7] = &EMPTYRAM;	break;
	case 1: Wm_blk[6] = &MAINRAM6;	Wm_blk[7] = &MAINRAM7;	break;
	case 2: Wm_blk[6] = &EXTRAM6;	Wm_blk[7] = &EXTRAM7;	break;
	case 3: Wm_blk[6] = &MAINRAM6;	Wm_blk[7] = &MAINRAM7;	break;	// Ƃ肠ɂĂ({͓OƂ)
	}
	
	// WX^ۑ
	Rf2 = data;
	
	#if (MEM_LOG)
	for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetWriteMemBlk(i) );
	#endif
}


////////////////////////////////////////////////////////////////
// ANZXEFCgݒ(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetWait( BYTE data )
{
	PRINTD( MEM_LOG, "[MEM][SetWait] -> M1:%d ROM:%d RAM:%d\n", (data>>7)&1, (data>>6)&1, (data>>5)&1 );
	
	// M1
	M1Wait = data&0x80 ? 1 : 0;
	
	// ROM
	int romwait = data&0x40 ? 1 : 0;
	EMPTYROM.SetWait( romwait, romwait );
	EXTROM0.SetWait( romwait, romwait );
	EXTROM1.SetWait( romwait, romwait );
	MAINROM0.SetWait( romwait, romwait );
	MAINROM1.SetWait( romwait, romwait );
	MAINROM2.SetWait( romwait, romwait );
	MAINROM3.SetWait( romwait, romwait );
	KANJIROM0.SetWait( romwait, romwait );
	KANJIROM1.SetWait( romwait, romwait );
	KANJIROM2.SetWait( romwait, romwait );
	KANJIROM3.SetWait( romwait, romwait );
	VOICEROM0.SetWait( romwait, romwait );
	VOICEROM1.SetWait( romwait, romwait );
	
	// RAM
	int ramwait = data&0x20 ? 1 : 0;
	EMPTYRAM.SetWait( ramwait, ramwait );
	MAINRAM0.SetWait( ramwait, ramwait );
	MAINRAM1.SetWait( ramwait, ramwait );
	MAINRAM2.SetWait( ramwait, ramwait );
	MAINRAM3.SetWait( ramwait, ramwait );
	MAINRAM4.SetWait( ramwait, ramwait );
	MAINRAM5.SetWait( ramwait, ramwait );
	MAINRAM6.SetWait( ramwait, ramwait );
	MAINRAM7.SetWait( ramwait, ramwait );
	EXTRAM0.SetWait( ramwait, ramwait );
	EXTRAM1.SetWait( ramwait, ramwait );
	EXTRAM2.SetWait( ramwait, ramwait );
	EXTRAM3.SetWait( ramwait, ramwait );
	EXTRAM4.SetWait( ramwait, ramwait );
	EXTRAM5.SetWait( ramwait, ramwait );
	EXTRAM6.SetWait( ramwait, ramwait );
	EXTRAM7.SetWait( ramwait, ramwait );
}


////////////////////////////////////////////////////////////////
// ANZXEFCg擾(62,66̂)
////////////////////////////////////////////////////////////////
BYTE MEM62::GetWait( void )
{
	return ( M1Wait ? 0x80 : 0 ) | ( EMPTYROM.GetWait() ? 0x40 : 0 ) | ( EMPTYRAM.GetWait() ? 0x20 : 0 );
}


////////////////////////////////////////////////////////////////
// ubNݒ(60̂)
////////////////////////////////////////////////////////////////
void MEM60::SetMemBlock( void )
{
	PRINTD( MEM_LOG, "[MEM][SetMemBlock]\n" );
	
	// Read
	Rm_blk[0] = &MAINROM0; Rm_blk[1] = &MAINROM1;
	Rm_blk[2] = pEXTROM0;  Rm_blk[3] = UseSol ? &SOLRAM : pEXTROM1;	// m̃J[gbWΉ
	Rm_blk[4] = &EXTRAM0;  Rm_blk[5] = &EXTRAM1;
	Rm_blk[6] = &MAINRAM0; Rm_blk[7] = &MAINRAM1;
	
	// Write
	Wm_blk[0] = &MAINROM0; Wm_blk[1] = &MAINROM1;
	Wm_blk[2] = pEXTROM0;  Wm_blk[3] = UseSol ? &SOLRAM : pEXTROM1; // m̃J[gbWΉ
	Wm_blk[4] = &EXTRAM0;  Wm_blk[5] = &EXTRAM1;
	Wm_blk[6] = &MAINRAM0; Wm_blk[7] = &MAINRAM1;
}


////////////////////////////////////////////////////////////////
// CG ROM AhXݒ(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SetCGrom( BYTE data )
{
	PRINTD( MEM_LOG, "[MEM][SetCGrom] -> %02x\n", data );
	
	// bit 7
	int	cgwait = data&0x80 ? 0 : 1;
	CGROM1.SetWait( cgwait, cgwait );
	CGROM2.SetWait( cgwait, cgwait );
	
	// bit 6
	cgenable = data&0x40 ? true : false;
	
	// bit 5,4,3
	// Ƃ肠
	
	// bit 2,1,0
	cgaddr   = data&3;
	
	SetMemBlockR( Rf0, Rf1 );
}


////////////////////////////////////////////////////////////////
// CG ROM I(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SelectCGrom( int mode )
{
	PRINTD( MEM_LOG, "[MEM][SelectCGrom] -> %d\n", mode );
	
	// mode 1:32*16(N60[h) 0:40*20(N60m[h)
	cgrom = mode ? true : false;
}


////////////////////////////////////////////////////////////////
// CG ROM BANK ؂ւ
////////////////////////////////////////////////////////////////
void MEM6::SetCGBank( bool data )
{
	PRINTD( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	Rm_blk[3] = CGBank ? &CGROM1 : UseSol ? &SOLRAM : pEXTROM1;
}

void MEM62::SetCGBank( bool data )
{
	PRINTD( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	SetMemBlockR( Rf0, Rf1 );
}


////////////////////////////////////////////////////////////////
// ROM  ROM ؂ւ(62,66̂)
////////////////////////////////////////////////////////////////
void MEM62::SelectKanjiRom( BYTE mode )
{
	PRINTD( MEM_LOG, "[MEM][SelectKanjiRom] -> %02X\n", mode );
	
// mode bit0 0:ROMI 1:ROMI
//      bit1 0:ROM     1:ROME

	//  HEH
	kj_LR = mode&0x02 ? true : false;
	
	// ROMHROMH
	kj_rom = mode&0x01 ? true : false;
	
	SetMemBlockR( Rf0, Rf1 );
}

void MEM66::SelectKanjiRom( BYTE mode )
{
	MEM62::SelectKanjiRom( mode );
	
	// 66̏ꍇ,ƈႤ炵
	if( !(mode&0x01) ){
		if( Rm_blk[0] != &MAINROM0 ) Rm_blk[0] = &VOICEROM0;
		if( Rm_blk[1] != &MAINROM1 ) Rm_blk[1] = &VOICEROM1;
	}
}


////////////////////////////////////////////////////////////////
// m̃J[gbW ROMoNݒ
////////////////////////////////////////////////////////////////
void MEM6::SetSolBank( BYTE data )
{
	SolBank = data & 0xf;
	EXTROM0.SetMemory( EXTROM0.GetName(), ExtRom + 0x2000 * SolBank, MemTable.ExtRom->WaitR, MemTable.ExtRom->WaitW, true );
	
	// RAM̃CgveNg
	SOLRAM.SetProtect( data & 0x10 ? true : false );
}


////////////////////////////////////////////////////////////////
// ڃANZX֐
////////////////////////////////////////////////////////////////
BYTE MEM6::ReadMainRom( WORD addr ){ return MainRom[addr&0x3fff]; }
BYTE MEM6::ReadMainRam( WORD addr ){ return IntRam[addr&0x3fff]; }
BYTE MEM6::ReadExtRom ( WORD addr ){ return ExtRom ? ExtRom[addr&0x3fff] : 0xff; }
BYTE MEM6::ReadCGrom1 ( WORD addr ){ return CGRom1[addr&0x1fff]; }
BYTE MEM6::ReadCGrom2 ( WORD )     { return 0xff; }

BYTE MEM60::ReadMainRam( WORD addr ){ return addr<0x4000 ? ExtRam[addr&0x3fff] : IntRam[addr&0x3fff]; }

BYTE MEM62::ReadMainRom( WORD addr ) { return MainRom[addr&0x7fff]; }
BYTE MEM62::ReadMainRam( WORD addr ) { return IntRam[addr&0xffff]; }
BYTE MEM62::ReadCGrom2( WORD addr )  { return CGRom2[addr&0x1fff]; }
BYTE MEM62::ReadKanjiRom( WORD addr ){ return KanjiRom[addr&0x7fff]; }
BYTE MEM62::ReadVoiceRom( WORD addr ){ return VoiceRom[addr&0x3fff]; }


////////////////////////////////////////////////////////////////
// I/OANZX֐
////////////////////////////////////////////////////////////////
void MEM6::Out7FH( int, BYTE data ){ SetSolBank( data ); }

void MEM62::OutC1H( int, BYTE data ){ SelectCGrom( (data>>1)&1 ); }
void MEM62::OutC2H( int, BYTE data ){ SelectKanjiRom( data ); }
void MEM62::OutC3H( int, BYTE data ){}
void MEM62::OutF0H( int, BYTE data ){ SetMemBlockR( data, Rf1 ); }
void MEM62::OutF1H( int, BYTE data ){ SetMemBlockR( Rf0, data ); }
void MEM62::OutF2H( int, BYTE data ){ SetMemBlockW( data ); }
void MEM62::OutF3H( int, BYTE data ){ SetWait( data ); }
void MEM62::OutF8H( int, BYTE data ){ SetCGrom( data ); }
BYTE MEM62::InF0H( int ){ return Rf0; }
BYTE MEM62::InF1H( int ){ return Rf1; }
BYTE MEM62::InF2H( int ){ return Rf2; }
BYTE MEM62::InF3H( int ){ return GetWait() | 0x1f; }


////////////////////////////////////////////////////////////////
// ǂłSAVE
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool MEM6::DokoSave( cIni *Ini )
{
	if( !Ini ) return false;
	
	Ini->PutEntry( "MEMORY", NULL, "CGBank",	 "%s",	CGBank ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "UseExtRam",	 "%s",	UseExtRam ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "M1Wait",	 "%d",	M1Wait );
	Ini->PutEntry( "MEMORY", NULL, "UseSoldier", "%s",	UseSol ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "SolBank",	 "%d",	SolBank );
	
	// gROM}EgĂꍇ
	if( UseExtRom )
		Ini->PutEntry( "MEMORY", NULL, "FilePath",	"%s",	FilePath );
	
	return true;
}

bool MEM60::DokoSave( cIni *Ini )
{
	PRINTD( MEM_LOG, "[MEM][DokoSave]\n" );
	
	if( !Ini || !MEM6::DokoSave( Ini ) ) return false;
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", IntRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	// ORAM
	if( UseExtRam || UseSol ){
		for( int i=0; i<0x4000; i+=64 ){		// 16KB
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
		}
	}
	
	if( UseSol ){
		// m̃J[gbWpORAM0x6000-0x7FFFgp
		for( int i=0x6000; i<0x8000; i+=64 ){	// 2KB
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
		}
	}
	
	return true;
}

bool MEM62::DokoSave( cIni *Ini )
{
	PRINTD( MEM_LOG, "[MEM][DokoSave]\n" );
	
	if( !Ini || !MEM6::DokoSave( Ini ) ) return false;
	
	Ini->PutEntry( "MEMORY", NULL, "cgrom",		 "%s",	cgrom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_rom",	 "%s",	kj_rom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_LR",		 "%s",	kj_LR ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "Rf0",	 "0x%02X",	Rf0 );
	Ini->PutEntry( "MEMORY", NULL, "Rf1",	 "0x%02X",	Rf1 );
	Ini->PutEntry( "MEMORY", NULL, "Rf2",	 "0x%02X",	Rf2 );
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", IntRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	// ORAM
	if( UseExtRam ){
		for( int i=0; i<0x10000; i+=64 ){		// 64KB
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
		}
	}
	
	if( UseSol ){
		// ݂̎ł͊ORAMgĂȂ̂Ŗ
		// m̃J[gbWpORAM0x6000-0x7FFFgp
//		for( int i=0x6000; i<0x8000; i+=64 ){	// 2KB
//			char stren[16],strva[256];
//			sprintf( stren, "ExtRam_%04X", i );
//			for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", ExtRam[i+j] );
//			Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
//		}
	}
	
	return true;
}



////////////////////////////////////////////////////////////////
// ǂłLOAD
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool MEM6::DokoLoad( cIni *Ini )
{
	char strfp[PATH_MAX];
	
	if( !Ini ) return false;
	
	Ini->GetTruth( "MEMORY", "CGBank",		&CGBank,	CGBank );
	Ini->GetTruth( "MEMORY", "UseExtRam",	&UseExtRam,	UseExtRam );
	Ini->GetInt(   "MEMORY", "M1Wait",		&M1Wait,	M1Wait );
	Ini->GetTruth( "MEMORY", "UseSoldier",	&UseSol,	UseSol );
	Ini->GetInt(   "MEMORY", "SolBank",		&SolBank,	SolBank );
	
	// gROM
	if( Ini->GetString( "MEMORY", "FilePath", strfp, "" ) )
		MountExtRom( strfp );
	
	return true;
}

bool MEM60::DokoLoad( cIni *Ini )
{
	PRINTD( MEM_LOG, "[MEM][DokoLoad]\n" );
	
	if( !Ini || !MEM6::DokoLoad( Ini ) ) return false;
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( &dt[2], &strva[j*2], 2 );
				IntRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// ORAM
	if( UseExtRam || UseSol ){
		for( int i=0; i<0x4000; i+=64 ){		// 16KB
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			memset( strva, '0', 64*2 );
			if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
				for( int j=0; j<64; j++ ){
					char dt[5] = "0x";
					strncpy( &dt[2], &strva[j*2], 2 );
					ExtRam[i+j] = strtol( dt, NULL, 16 );
				}
			}
		}
	}
	
	if( UseSol ){
		// m̃J[gbWpORAM0x6000-0x7FFFgp
		for( int i=0x6000; i<0x8000; i+=64 ){	// 2KB
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			memset( strva, '0', 64*2 );
			if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
				for( int j=0; j<64; j++ ){
					char dt[5] = "0x";
					strncpy( &dt[2], &strva[j*2], 2 );
					ExtRam[i+j] = strtol( dt, NULL, 16 );
				}
			}
		}
	}
	
	Init( UseExtRam, UseSol );
	SetMemBlock();
	SetCGBank( CGBank );
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( SolBank );	// oN
	// -------------------------------------------------------------
	
	return true;
}

bool MEM62::DokoLoad( cIni *Ini )
{
	PRINTD( MEM_LOG, "[MEM][DokoLoad]\n" );
	
	int st;
	
	if( !Ini || !MEM6::DokoLoad( Ini ) ) return false;
	
	Ini->GetTruth( "MEMORY", "cgrom",		&cgrom,		cgrom );
	Ini->GetTruth( "MEMORY", "kj_rom",		&kj_rom,	kj_rom );
	Ini->GetTruth( "MEMORY", "kj_LR",		&kj_LR,		kj_LR );
	Ini->GetInt(   "MEMORY", "Rf0",			&st,		Rf0 );	Rf0 = st;
	Ini->GetInt(   "MEMORY", "Rf1",			&st,		Rf1 );	Rf1 = st;
	Ini->GetInt(   "MEMORY", "Rf2",			&st,		Rf2 );	Rf2 = st;
	
	// RAM
	for( int i=0; i<(int)IINTRAM.Size; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "IntRam_%04X", i );
		memset( strva, '0', 64*2 );
		if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
			for( int j=0; j<64; j++ ){
				char dt[5] = "0x";
				strncpy( &dt[2], &strva[j*2], 2 );
				IntRam[i+j] = strtol( dt, NULL, 16 );
			}
		}
	}
	
	// ORAM
	if( UseExtRam ){
		for( int i=0; i<0x10000; i+=64 ){
			char stren[16],strva[256];
			sprintf( stren, "ExtRam_%04X", i );
			memset( strva, '0', 64*2 );
			if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
				for( int j=0; j<64; j++ ){
					char dt[5] = "0x";
					strncpy( dt, &strva[j*2], 2 );
					ExtRam[i+j] = strtol( dt, NULL, 16 );
				}
			}
		}
	}
	
	if( UseSol ){
		// ݂̎ł͊ORAMgĂȂ̂Ŗ
		// m̃J[gbWpORAM0x6000-0x7FFFgp
//		for( int i=0x6000; i<0x8000; i+=64 ){	// 2KB
//			char stren[16],strva[256];
//			sprintf( stren, "ExtRam_%04X", i );
//			memset( strva, '0', 64*2 );
//			if( Ini->GetString( "MEMORY", stren, strva, strva ) ){
//				for( int j=0; j<64; j++ ){
//					char dt[5] = "0x";
//					strncpy( &dt[2], &strva[j*2], 2 );
//					ExtRam[i+j] = strtol( dt, NULL, 16 );
//				}
//			}
//		}
	}
	
	Init( UseExtRam, UseSol );
	SetMemBlockR( Rf0, Rf1 );
	SetMemBlockW( Rf2 );
	SetCGBank( CGBank );
	
	// m̃J[gbW ------------------------------------------
	if( UseSol ) SetSolBank( SolBank );	// oN
	// -------------------------------------------------------------
	
	return true;
}


#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
////////////////////////////////////////////////////////////////
// fobOp֐
////////////////////////////////////////////////////////////////
const char *MEM6::GetReadMemBlk( int blk )
{
	return Rm_blk[blk]->GetName();
}

const char *MEM6::GetWriteMemBlk( int blk )
{
	return Wm_blk[blk]->GetName();
}
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //



////////////////////////////////////////////////////////////////
//  device description
////////////////////////////////////////////////////////////////
const Device::Descriptor MEM60::descriptor = {
	MEM60::indef, MEM60::outdef
};

const Device::OutFuncPtr MEM60::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &MEM6::Out7FH )
};

const Device::InFuncPtr MEM60::indef[] = { NULL };

const Device::Descriptor MEM62::descriptor = {
	MEM62::indef, MEM62::outdef
};

const Device::OutFuncPtr MEM62::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &MEM6::Out7FH ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC2H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC3H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF0H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF2H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF3H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF8H )
};

const Device::InFuncPtr MEM62::indef[] = {
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF0H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF1H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF2H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF3H )
};
