#include <new>

#include "pc6001v.h"

#include "breakpoint.h"
#include "config.h"
#include "cpum.h"
#include "cpus.h"
#include "debug.h"
#include "disk.h"
#include "graph.h"
#include "intr.h"
#include "io.h"
#include "keyboard.h"
#include "log.h"
#include "memory.h"
#include "pio.h"
#include "psg.h"
#include "schedule.h"
#include "sound.h"
#include "tape.h"
#include "voice.h"
#include "vdg.h"

#include "p6el.h"
#include "p6vm.h"



// P6foCXp֐Q
// EL ----------------------------------------------------------

////////////////////////////////////////////////////////////////
// j^[h?
////////////////////////////////////////////////////////////////
#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
bool VM6::ElIsMonitor( void )
{
	return el->IsMonitor();
}
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //




// EVSC --------------------------------------------------------

////////////////////////////////////////////////////////////////
// Cxgǉ
////////////////////////////////////////////////////////////////
bool VM6::EventAdd( Device *dev, int id, double hz, int flag )
{
	return evsc->Add( dev, id, hz, flag );
}


////////////////////////////////////////////////////////////////
// Cxg폜
////////////////////////////////////////////////////////////////
bool VM6::EventDel( Device *dev, int id )
{
	return evsc->Del( dev, id );
}


////////////////////////////////////////////////////////////////
// CxgXV
////////////////////////////////////////////////////////////////
void VM6::EventUpdate( int clk )
{
	evsc->Update( clk );
}


////////////////////////////////////////////////////////////////
// wCxgZbg
////////////////////////////////////////////////////////////////
void VM6::EventReset( Device::ID devid, int id, double ini )
{
	evsc->Reset( devid, id, ini );
}


////////////////////////////////////////////////////////////////
// Cxg̐is߂
////////////////////////////////////////////////////////////////
double VM6::EventScale( Device::ID devid, int id )
{
	return evsc->Scale( devid, id );
}


////////////////////////////////////////////////////////////////
// Cxg擾
////////////////////////////////////////////////////////////////
bool VM6::EventGetInfo( EVSC::evinfo *info )
{
	return evsc->GetEvinfo( info );
}


////////////////////////////////////////////////////////////////
// Cxgݒ
////////////////////////////////////////////////////////////////
bool VM6::EventSetInfo( EVSC::evinfo *info )
{
	return evsc->SetEvinfo( info );
}


////////////////////////////////////////////////////////////////
// VSYNCʒm
////////////////////////////////////////////////////////////////
void VM6::EventOnVSYNC( void )
{
	evsc->OnVSYNC();
}



// IO6 ---------------------------------------------------------

////////////////////////////////////////////////////////////////
// IN֐
////////////////////////////////////////////////////////////////
BYTE VM6::IomIn( int port, int *wcnt )
{
	BYTE data = iom->In( port, wcnt );
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg`FbN
	if( BpCheckBreakPoint( BPoint::BP_IN, port&0xff, data, NULL ) ){
		PRINTD( IO_LOG, " -> Break!\n" );
	}
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	return data;
}

BYTE VM6::IosIn( int port, int *wcnt )
{
	return ios->In( port, wcnt );
}


////////////////////////////////////////////////////////////////
// OUT֐
////////////////////////////////////////////////////////////////
void VM6::IomOut( int port, BYTE data, int *wcnt )
{
	iom->Out( port, data, wcnt );
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg`FbN
	if( BpCheckBreakPoint( BPoint::BP_OUT, port&0xff, data, NULL ) ){
		PRINTD( IO_LOG, " -> Break!\n" );
	}
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
}

void VM6::IosOut( int port, BYTE data, int *wcnt )
{
	ios->Out( port, data, wcnt );
}



// INT6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// ݃`FbN
////////////////////////////////////////////////////////////////
int VM6::IntIntrCheck( void )
{
	return intr->IntrCheck();
}


////////////////////////////////////////////////////////////////
// ݗv
////////////////////////////////////////////////////////////////
void VM6::IntReqIntr( DWORD vec )
{
	intr->ReqIntr( vec );
}


////////////////////////////////////////////////////////////////
// ݓP
////////////////////////////////////////////////////////////////
void VM6::IntCancelIntr( DWORD vec )
{
	intr->CancelIntr( vec );
}



// CPU6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// 1CtAZu
////////////////////////////////////////////////////////////////
int VM6::CpumDisasm( char *str, WORD pc )
{
	return cpum->Disasm( str, pc );
}


////////////////////////////////////////////////////////////////
// WX^l擾
////////////////////////////////////////////////////////////////
void VM6::CpumGetRegister( cZ80::Register * reg )
{
	cpum->GetRegister( reg );
}


////////////////////////////////////////////////////////////////
// WX^lݒ
////////////////////////////////////////////////////////////////
void VM6::CpumSetRegister( cZ80::Register * reg )
{
	cpum->SetRegister( reg );
}



// SUB6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// L[ݗv
////////////////////////////////////////////////////////////////
void VM6::CpusReqKeyIntr( int flag, BYTE data )
{
	cpus->ReqKeyIntr( flag, data );
}


////////////////////////////////////////////////////////////////
// Oݗv
////////////////////////////////////////////////////////////////
void VM6::CpusExtIntr( void )
{
	cpus->ExtIntr();
}


////////////////////////////////////////////////////////////////
// CMT READݗv
////////////////////////////////////////////////////////////////
void VM6::CpusReqCmtIntr( BYTE data )
{
	cpus->ReqCmtIntr( data );
}


////////////////////////////////////////////////////////////////
// CMTݔ?
////////////////////////////////////////////////////////////////
bool VM6::CpusIsCmtIntrReady( void )
{
	return cpus->IsCmtIntrReady();
}



// MEM6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// tFb`(M1)
////////////////////////////////////////////////////////////////
BYTE VM6::MemFetch( WORD addr, int *m1wait )
{
	BYTE data = mem->Fetch( addr, m1wait );
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg`FbN
	if( BpCheckBreakPoint( BPoint::BP_READ, addr, data, NULL ) ){
		PRINTD( MEM_LOG, " -> Break!\n" );
	}
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	return data;
}


////////////////////////////////////////////////////////////////
// [h
////////////////////////////////////////////////////////////////
BYTE VM6::MemRead( WORD addr, int *wcnt )
{
	BYTE data = mem->Read( addr, wcnt );
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg`FbN
	if( BpCheckBreakPoint( BPoint::BP_READ, addr, data, NULL ) ){
		PRINTD( MEM_LOG, " -> Break!\n" );
	}
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	return data;
}


////////////////////////////////////////////////////////////////
// Cg
////////////////////////////////////////////////////////////////
void VM6::MemWrite( WORD addr, BYTE data, int *wcnt )
{
	mem->Write( addr, data, wcnt );
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg`FbN
	if( BpCheckBreakPoint( BPoint::BP_WRITE, addr, data, NULL ) ){
		PRINTD( MEM_LOG, " -> Break!\n" );
	}
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
}

////////////////////////////////////////////////////////////////
// CG ROM BANK ؂ւ
////////////////////////////////////////////////////////////////
void VM6::MemSetCGBank( bool data )
{
	mem->SetCGBank( data );
}

BYTE VM6::MemReadMainRom( WORD addr ){ return mem->ReadMainRom( addr ); }
BYTE VM6::MemReadMainRam( WORD addr ){ return mem->ReadMainRam( addr ); }
BYTE VM6::MemReadExtRom ( WORD addr ){ return mem->ReadExtRom( addr ); }
BYTE VM6::MemReadCGrom1 ( WORD addr ){ return mem->ReadCGrom1( addr ); }
BYTE VM6::MemReadCGrom2 ( WORD addr ){ return mem->ReadCGrom2( addr ); }
#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
const char *VM6::MemGetReadMemBlk( int blk ) { return mem->GetReadMemBlk( blk );  }
const char *VM6::MemGetWriteMemBlk( int blk ){ return mem->GetWriteMemBlk( blk ); }
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //



// VDG ---------------------------------------------------------

////////////////////////////////////////////////////////////////
// CRT\Ԑݒ
////////////////////////////////////////////////////////////////
void VM6::VdgSetCrtDisp( bool st )
{
	vdg->SetCrtDisp( st );
}


////////////////////////////////////////////////////////////////
// EBhETCY擾
////////////////////////////////////////////////////////////////
bool VM6::VdgGetWinSize( void )
{
	return vdg->GetWinSize();
}


////////////////////////////////////////////////////////////////
// \ԃtO擾
////////////////////////////////////////////////////////////////
bool VM6::VdgIsDisp( void )
{
	return vdg->IsDisp();
}



// PIO6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// PartA [h
////////////////////////////////////////////////////////////////
BYTE VM6::PioReadA( void )
{
	return pio->ReadA();
}



// KEY6 --------------------------------------------------------

////////////////////////////////////////////////////////////////
// J[\L[Ԏ擾
////////////////////////////////////////////////////////////////
BYTE VM6::KeyGetKeyJoy( void )
{
	return key->GetKeyJoy();
}


////////////////////////////////////////////////////////////////
// WCXeBbNԎ擾
////////////////////////////////////////////////////////////////
BYTE VM6::KeyGetJoy( int JoyNo )
{
	return key->GetJoy( JoyNo );
}


////////////////////////////////////////////////////////////////
// L[{[hCWP[^Ԏ擾
////////////////////////////////////////////////////////////////
BYTE VM6::KeyGetKeyIndicator( void )
{
	return key->GetKeyIndicator();
}


////////////////////////////////////////////////////////////////
// p<->Ȑ؊
////////////////////////////////////////////////////////////////
void VM6::KeyChangeKana( void )
{
	key->ChangeKana();
}


////////////////////////////////////////////////////////////////
// <->Ji؊
////////////////////////////////////////////////////////////////
void VM6::KeyChangeKKana( void )
{
	key->ChangeKKana();
}



// CMTL --------------------------------------------------------

////////////////////////////////////////////////////////////////
// }Egς?
////////////////////////////////////////////////////////////////
bool VM6::CmtlIsMount( void )
{
	return cmtl->IsMount();
}


////////////////////////////////////////////////////////////////
// I[gX^[g?
////////////////////////////////////////////////////////////////
bool VM6::CmtlIsAutoStart( void )
{
	return cmtl->IsAutoStart();
}


////////////////////////////////////////////////////////////////
// BoostUpԎ擾
////////////////////////////////////////////////////////////////
bool VM6::CmtlIsBoostUp( void )
{
	return cmtl->IsBoostUp();
}


////////////////////////////////////////////////////////////////
// t@CpX擾
////////////////////////////////////////////////////////////////
const char *VM6::CmtlGetFile()
{
	return cmtl->GetFile();
}


////////////////////////////////////////////////////////////////
// TAPE擾
////////////////////////////////////////////////////////////////
const char *VM6::CmtlGetName( void )
{
	return cmtl->GetName();
}


////////////////////////////////////////////////////////////////
// x^C[WTCY擾
////////////////////////////////////////////////////////////////
DWORD VM6::CmtlGetSize( void )
{
	return cmtl->GetSize();
}


////////////////////////////////////////////////////////////////
// JE^擾
////////////////////////////////////////////////////////////////
int VM6::CmtlGetCount( void )
{
	return cmtl->GetCount();
}



// CMTS --------------------------------------------------------

////////////////////////////////////////////////////////////////
// TAPE}Eg
////////////////////////////////////////////////////////////////
bool VM6::CmtsMount( void )
{
	return cmts->Mount();
}


////////////////////////////////////////////////////////////////
// TAPEA}Eg
////////////////////////////////////////////////////////////////
void VM6::CmtsUnmount( void )
{
	cmts->Unmount();
}


////////////////////////////////////////////////////////////////
// {[[gݒ
////////////////////////////////////////////////////////////////
void VM6::CmtsSetBaud( int b )
{
	cmts->SetBaud( b );
}


////////////////////////////////////////////////////////////////
// CMT 1
////////////////////////////////////////////////////////////////
void VM6::CmtsCmtWrite( BYTE data )
{
	cmts->CmtWrite( data );
}



// DSK ---------------------------------------------------------

////////////////////////////////////////////////////////////////
// }Egς?
////////////////////////////////////////////////////////////////
bool VM6::DskIsMount( int drvno )
{
	return disk->IsMount( drvno );
}


////////////////////////////////////////////////////////////////
// VXefBXN?
////////////////////////////////////////////////////////////////
bool VM6::DskIsSystem( int drvno )
{
	return disk->IsSystem( drvno );
}


////////////////////////////////////////////////////////////////
// veNg?
////////////////////////////////////////////////////////////////
bool VM6::DskIsProtect( int drvno )
{
	return disk->IsProtect( drvno );
}


////////////////////////////////////////////////////////////////
// ANZX?
////////////////////////////////////////////////////////////////
bool VM6::DskInAccess( int drvno )
{
	return disk->InAccess( drvno );
}


////////////////////////////////////////////////////////////////
// t@CpX擾
////////////////////////////////////////////////////////////////
const char *VM6::DskGetFile( int drvno )
{
	return disk->GetFile( drvno );
}


////////////////////////////////////////////////////////////////
// DISK擾
////////////////////////////////////////////////////////////////
const char *VM6::DskGetName( int drvno )
{
	return disk->GetName( drvno );
}



// BPoint ------------------------------------------------------

////////////////////////////////////////////////////////////////
// u[N|Cg̗L`FbN
////////////////////////////////////////////////////////////////
bool VM6::BpExistBreakPoint( void )
{
	return bp->ExistBreakPoint();
}


////////////////////////////////////////////////////////////////
// u[N|Cg`FbN
////////////////////////////////////////////////////////////////
bool VM6::BpCheckBreakPoint( BPoint::BPtype type, WORD addr, BYTE data, const char *str )
{
	return bp->CheckBreakPoint( type, addr, data, str );
}


////////////////////////////////////////////////////////////////
// u[N|Cg̏NA
////////////////////////////////////////////////////////////////
void VM6::BpClearStatus( void )
{
	bp->ClearStatus();
}


////////////////////////////////////////////////////////////////
// u[N|Cg̃^Cv擾
////////////////////////////////////////////////////////////////
BPoint::BPtype VM6::BpGetType( int num )
{
	return bp->GetType( num );
}


////////////////////////////////////////////////////////////////
// u[N|Cg̃^Cvݒ
////////////////////////////////////////////////////////////////
void VM6::BpSetType( int num, BPoint::BPtype type )
{
	bp->SetType( num, type );
}


////////////////////////////////////////////////////////////////
// u[N|Cg̃AhX擾
////////////////////////////////////////////////////////////////
WORD VM6::BpGetAddr( int num )
{
	return bp->GetAddr( num );
}


////////////////////////////////////////////////////////////////
// u[N|Cg̃AhXݒ
////////////////////////////////////////////////////////////////
void VM6::BpSetAddr( int num, WORD addr )
{
	bp->SetAddr( num, addr );
}


////////////////////////////////////////////////////////////////
// u[Nv?
////////////////////////////////////////////////////////////////
bool VM6::BpIsReqBreak( void )
{
	return bp->IsReqBreak();
}


////////////////////////////////////////////////////////////////
// u[Nv̂u[N|CgNo.擾
////////////////////////////////////////////////////////////////
int VM6::BpGetReqBPNum( void )
{
	return bp->GetReqBPNum();
}


////////////////////////////////////////////////////////////////
// u[NvLZ
////////////////////////////////////////////////////////////////
void VM6::BpResetBreak( void )
{
	bp->ResetBreak();
}








////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
VM6::VM6( EL6 *emuobj ) : clock(0), el(emuobj), evsc(NULL), iom(NULL), ios(NULL),
	intr(NULL), cpum(NULL), cpus(NULL), mem(NULL),
	vdg(NULL), psg(NULL), voice(NULL), pio(NULL), key(NULL),
	cmtl(NULL), cmts(NULL), disk(NULL)
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	, bp(NULL)
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	{}

VM60::VM60( EL6 *emuobj ) : VM6(emuobj)
{
	clock = CPUM_CLOCK60;
	
	DevTable.Intr    = VM60::c_intr;	// 
	DevTable.Vdg     = VM60::c_vdg;		// VDG
	DevTable.Psg     = VM60::c_psg;		// PSG
	DevTable.M8255   = VM60::c_8255m;	// I/O(Z80)
	DevTable.S8255   = VM60::c_8255s;	// I/O(SUB CPU)
	DevTable.Disk    = VM60::c_disk;	// DISK
	DevTable.CmtL    = VM60::c_cmtl;	// CMT(LOAD)
	
	DevTable.Soldier = VM6::c_soldier;	// m̃J[gbW
}

VM61::VM61( EL6 *emuobj ) : VM6(emuobj)
{
	clock = CPUM_CLOCK60;
	
	DevTable.Intr    = VM60::c_intr;	// 
	DevTable.Vdg     = VM60::c_vdg;		// VDG
	DevTable.Psg     = VM60::c_psg;		// PSG
	DevTable.M8255   = VM60::c_8255m;	// I/O(Z80)
	DevTable.S8255   = VM60::c_8255s;	// I/O(SUB CPU)
	DevTable.Disk    = VM60::c_disk;	// DISK
	DevTable.CmtL    = VM60::c_cmtl;	// CMT(LOAD)
	
	DevTable.Soldier = VM6::c_soldier;	// m̃J[gbW
}

VM62::VM62( EL6 *emuobj ) : VM6(emuobj)
{
	clock = CPUM_CLOCK62;
	
	DevTable.Intr    = VM62::c_intr;	// 
	DevTable.Memory  = VM62::c_memory;	// 
	DevTable.Vdg     = VM62::c_vdg;		// VDG
	DevTable.Psg     = VM62::c_psg;		// PSG
	DevTable.M8255   = VM62::c_8255m;	// I/O(Z80)
	DevTable.S8255   = VM62::c_8255s;	// I/O(SUB CPU)
	DevTable.Voice   = VM62::c_voice;	// 
	DevTable.Disk    = VM62::c_disk;	// DISK
	DevTable.CmtL    = VM62::c_cmtl;	// CMT(LOAD)
	
	DevTable.Soldier = VM6::c_soldier;	// m̃J[gbW
}

VM66::VM66( EL6 *emuobj ) : VM6(emuobj)
{
	clock = CPUM_CLOCK62;
	
	DevTable.Intr    = VM62::c_intr;	// 
	DevTable.Memory  = VM62::c_memory;	// 
	DevTable.Vdg     = VM62::c_vdg;		// VDG
	DevTable.Psg     = VM62::c_psg;		// PSG
	DevTable.M8255   = VM62::c_8255m;	// I/O(Z80)
	DevTable.S8255   = VM62::c_8255s;	// I/O(SUB CPU)
	DevTable.Voice   = VM62::c_voice;	// 
	DevTable.Disk    = VM66::c_disk;	// DISK
	DevTable.CmtL    = VM62::c_cmtl;	// CMT(LOAD)
	
	DevTable.Soldier = VM6::c_soldier;	// m̃J[gbW
}



////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
VM6::~VM6( void )
{
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	if( bp )     delete bp;
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	if( voice )  delete voice;
	if( key )    delete key;
	if( pio )    delete pio;
	if( cpus )   delete cpus;
	if( cmts )   delete cmts;
	if( cmtl )   delete cmtl;
	if( disk )   delete disk;
	if( cpum )   delete cpum;
	if( intr )   delete intr;
	if( vdg )    delete vdg;
	if( mem )    delete mem;
	if( psg )    delete psg;
	if( ios )    delete ios;
	if( iom )    delete iom;
	if( evsc )   delete evsc;
}

VM60::~VM60( void ){}

VM61::~VM61( void ){}

VM62::~VM62( void ){}

VM66::~VM66( void ){}


////////////////////////////////////////////////////////////////
// @ʃIuWFNgm
////////////////////////////////////////////////////////////////
bool VM60::AllocObjSpecific( void )
{
	cpus   = new SUB60( this, DEV_ID("8049") );		// SUB CPU
	intr   = new INT60( this, DEV_ID("INTR") );		// 
	mem    = new MEM60( this, DEV_ID("MEM1") );		// 
	vdg    = new VDG60( this, DEV_ID("VDG1") );		// VDG
	key    = new KEY60( this, DEV_ID("KEYB") );		// L[
	disk   = new DSK60( this, DEV_ID("DISK") );		// DISK
	
	return true;
}

bool VM61::AllocObjSpecific( void )
{
	cpus   = new SUB60( this, DEV_ID("8049") );		// SUB CPU
	intr   = new INT60( this, DEV_ID("INTR") );		// 
	mem    = new MEM61( this, DEV_ID("MEM1") );		// 
	vdg    = new VDG60( this, DEV_ID("VDG1") );		// VDG
	key    = new KEY60( this, DEV_ID("KEYB") );		// L[
	disk   = new DSK60( this, DEV_ID("DISK") );		// DISK
	
	return true;
}

bool VM62::AllocObjSpecific( void )
{
	cpus   = new SUB62( this, DEV_ID("8049") );		// SUB CPU
	intr   = new INT62( this, DEV_ID("INTR") );		// 
	mem    = new MEM62( this, DEV_ID("MEM1") );		// 
	vdg    = new VDG62( this, DEV_ID("VDG2") );		// VDG
	key    = new KEY6 ( this, DEV_ID("KEYB") );		// L[
	voice  = new VCE6 ( this, DEV_ID("VOIC") );		// 
	disk   = new DSK60( this, DEV_ID("DISK") );		// DISK
	
	return true;
}

bool VM66::AllocObjSpecific( void )
{
	cpus   = new SUB62( this, DEV_ID("8049") );		// SUB CPU
	intr   = new INT62( this, DEV_ID("INTR") );		// 
	mem    = new MEM66( this, DEV_ID("MEM1") );		// 
	vdg    = new VDG62( this, DEV_ID("VDG2") );		// VDG
	key    = new KEY6 ( this, DEV_ID("KEYB") );		// L[
	voice  = new VCE6 ( this, DEV_ID("VOIC") );		// 
	disk   = new DSK66( this, DEV_ID("DISK") );		// DISK
	
	return true;
}


////////////////////////////////////////////////////////////////
// SIuWFNgm
////////////////////////////////////////////////////////////////
bool VM6::AllocObject( CFG6 *cnfg )
{
	PRINTD( VM_LOG, "[VM][AllocObject]\n" );
	
	// eIuWFNgm
	try{
		evsc   = new EVSC( clock ) ;				// CxgXPW[
		iom    = new IO6;							// I/O(Z80)
		ios    = new IO6;							// I/O(SUB CPU)
		cpum   = new CPU6 ( this, DEV_ID("CPU1") );	// CPU
		psg    = new PSG6 ( this, DEV_ID("PSG1") );	// PSG
		pio    = new PIO6 ( this, DEV_ID("8255") );	// 8255
		cmtl   = new CMTL ( this, DEV_ID("TAPE") );	// CMT(LOAD)
		cmts   = new CMTS ( this, DEV_ID("SAVE") );	// CMT(SAVE)
		#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		bp     = new BPoint();						// u[N|Cg
		#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	}
	// new Ɏsꍇ
	catch( std::bad_alloc ){
		if( evsc ){ delete evsc; evsc = NULL; }
		if( iom ) { delete iom;  iom  = NULL; }
		if( ios ) { delete ios;  ios  = NULL; }
		if( cpum ){ delete cpum; cpum = NULL; }
		if( psg ) { delete psg;  psg  = NULL; }
		if( pio ) { delete pio;  pio  = NULL; }
		if( cmtl ){ delete cmtl; cmtl = NULL; }
		if( cmts ){ delete cmts; cmts = NULL; }
		#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		if( bp )  { delete bp;   bp   = NULL; }
		#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		return false;
	}
	
	// @ʃIuWFNgm
	if( !AllocObjSpecific() ){
		if( cpus ) { delete cpus;  cpus  = NULL; }
		if( intr ) { delete intr;  intr  = NULL; }
		if( mem )  { delete mem;   mem   = NULL; }
		if( vdg )  { delete vdg;   vdg   = NULL; }
		if( key )  { delete key;   key   = NULL; }
		if( voice ){ delete voice; voice = NULL; }
		if( disk ) { delete disk;  disk  = NULL; }
		return false;
	}
	
	// SmۂROMt@CǍ
	if( !mem->AllocAllMemory( cnfg->GetRomPath(), cnfg->GetCheckCRC() ) ) return false;
	if( cnfg->GetRomPatch() ) mem->Patch();
	
	return true;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
bool VM6::Init( CFG6 *cnfg  )
{
	PRINTD( VM_LOG, "[VM][Init]\n" );
	
	// CxgXPW[
	evsc->SetMasterClock( clock * cnfg->GetOverClock() / 100 );
	evsc->Entry( intr );
	evsc->Entry( cpum );
	evsc->Entry( cpus );
	evsc->Entry( vdg );
	evsc->Entry( psg );
	evsc->Entry( voice );
	evsc->Entry( pio );
	evsc->Entry( key );
	evsc->Entry( cmtl );
	evsc->Entry( cmts );
	evsc->Entry( disk );
	
	// I/O(Z80)@-----
	if( !iom->Init( 256 ) ) return false;
	
	// I/O(SUB CPU)@-----
	if( !ios->Init( 10 ) ) return false;
	
	//  -----
	intr->Reset();
	
	// CPU -----
	cpum->Reset();
	
	// SUB CPU -----
	cpus->Reset();
	
	//  -----
	if( !mem->Init( cnfg->GetUseExtRam(), cnfg->GetUseSoldier() ) ) return false;
	mem->Reset();
	if( *(cnfg->GetExtRomFile()) ) if( !mem->MountExtRom( cnfg->GetExtRomFile() ) ) return false;
	
	// VDG -----
	if( !vdg->Init() ) return false;
	vdg->SetMode4Color( cnfg->GetMode4Color() );
	
	// PSG -----
	psg->SetVolume( cnfg->GetPsgVol() );		// ʐݒ
	psg->SetLPF( cnfg->GetPsgLPF() );			// [pXtB^ JbgItgݒ
	int i=0;
	while( DevTable.Psg[i].bank ){						// EFCgݒ
		if( DevTable.Psg[i].rule == IOBus::portout ) iom->SetOutWait( DevTable.Psg[i].bank, 1 );
		else										 iom->SetInWait ( DevTable.Psg[i].bank, 1 );
		i++;
	}
	if( !psg->Init( clock/2, cnfg->GetSampleRate() ) ) return false;
	
	
	// 8255 -----
	pio->Reset();
	pio->cPRT::Init( cnfg->GetPrinterFile() );
	
	// L[ -----
	if( !key->Init( cnfg->GetKeyRepeat() ) ) return false;
	VKeyConv *vk;
	if( cnfg->GetVKeyDef( &vk ) )				// L[`z擾
		key->SetVKeySymbols( vk );				// zL[R[h -> P6L[R[h ݒ
	
	// CMT(LOAD) -----
	if( !cmtl->Init( cnfg->GetSampleRate() ) ) return false;
	cmtl->SetVolume( cnfg->GetCmtVol() );		// ʐݒ
	cmtl->SetLPF( DEFAULT_TAPELPF );			// [pXtB^ JbgItgݒ
	cmtl->SetBoost( cnfg->GetBoostUp() );
	cmtl->SetMaxBoost( cnfg->GetMaxBoost1(), cnfg->GetMaxBoost2() );
	
	// CMT(SAVE) -----
	if( !cmts->Init( cnfg->GetSaveFile() ) ) return false;
	
	// DISK -----
//	if( cnfg->GetFddNum() )
		if( !disk->Init( cnfg->GetFddNum() ) ) return false;
	
	//  -----
	if( DevTable.Voice ){
		if( !voice->Init( cnfg->GetSampleRate(), cnfg->GetWavePath() ) ) return false;
		voice->SetVolume( cnfg->GetVoiceVol() );	// ʐݒ
	}
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// u[N|Cg
	bp->ClearStatus();
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	
	
	
	// I/O|[gɃfoCXڑ
	if( !iom->Connect( intr, DevTable.Intr  ) ) return false;	// 
	if( !iom->Connect( vdg,  DevTable.Vdg   ) ) return false;	// VDG
	if( !iom->Connect( psg,  DevTable.Psg   ) ) return false;	// PSG
	if( !iom->Connect( pio,  DevTable.M8255 ) ) return false;	// 8255(Z80)
	if( !ios->Connect( pio,  DevTable.S8255 ) ) return false;	// 8255(SUB CPU)
	if( !iom->Connect( cmtl, DevTable.CmtL  ) ) return false;	// CMT(LOAD)
	if( cnfg->GetFddNum() || (cnfg->GetModel() == 66) )			// DISK
		if( !iom->Connect( disk,  DevTable.Disk   ) ) return false;
	if( DevTable.Memory )										// 
		if( !iom->Connect( mem,   DevTable.Memory ) ) return false;
	if( DevTable.Voice )										// 
		if( !iom->Connect( voice, DevTable.Voice  ) ) return false;
	
	// IvV@\ -----
	if( cnfg->GetUseSoldier() )									// m̃J[gbW
		if( !iom->Connect( mem, DevTable.Soldier ) ) return false;
	
	
	return true;
}


////////////////////////////////////////////////////////////////
// Zbg
////////////////////////////////////////////////////////////////
void VM6::Reset( void )
{
	PRINTD( VM_LOG, "[VM][Reset]\n" );
	
	intr->Reset();
	cpum->Reset();
	cpus->Reset();
	mem->Reset();
	psg->Reset();
	pio->Reset();
	cmtl->Reset();
	disk->Reset();
	if( voice ) voice->Reset();
}


////////////////////////////////////////////////////////////////
// 1ߎs
////////////////////////////////////////////////////////////////
int VM6::Emu( void )
{
	PRINTD( VM_LOG, "[VM][Emu]\n" );
	
	return cpum->Exec();	// CPU 1ߎs
}


////////////////////////////////////////////////////////////////
// CPUNbN擾
////////////////////////////////////////////////////////////////
int VM6::GetCPUClock( void )
{
	PRINTD( VM_LOG, "[VM][GetCPUClock]\n" );
	
	return clock;
}








////////////////////////////////////////////////////////////////
// foCXRlN^
////////////////////////////////////////////////////////////////

// m̃J[gbW
const IOBus::Connector VM6::c_soldier[] = {
	{ 0x7f, IOBus::portout, MEM6::out7FH },
	{ 0, 0, 0 }	};



// PC-6001,PC-6001A --------------------------------------------

//  -----
const IOBus::Connector VM60::c_intr[] = {
	{ 0xb0, IOBus::portout, INT60::outB0H },
	{ 0xb1, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb2, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb3, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb4, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb5, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb6, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb7, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb8, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xb9, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xba, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xbb, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xbc, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xbd, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xbe, IOBus::portout, INT60::outB0H },	// C[W
	{ 0xbf, IOBus::portout, INT60::outB0H },	// C[W
	{ 0, 0, 0 }	};

// VDG -----
const IOBus::Connector VM60::c_vdg[] = {
	{ 0xb0, IOBus::portout, VDG60::outB0H },
	{ 0xb1, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb2, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb3, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb4, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb5, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb6, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb7, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb8, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xb9, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xba, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xbb, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xbc, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xbd, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xbe, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0xbf, IOBus::portout, VDG60::outB0H },	// C[W
	{ 0, 0, 0 } };

// PSG -----
const IOBus::Connector VM60::c_psg[] = {
	{ 0xa0, IOBus::portout, PSG6::outA0H },
	{ 0xa1, IOBus::portout, PSG6::outA1H },
	{ 0xa3, IOBus::portout, PSG6::outA3H },
	{ 0xa4, IOBus::portout, PSG6::outA0H },	// C[W
	{ 0xa5, IOBus::portout, PSG6::outA1H },	// C[W
	{ 0xa7, IOBus::portout, PSG6::outA3H },	// C[W
	{ 0xa8, IOBus::portout, PSG6::outA0H },	// C[W
	{ 0xa9, IOBus::portout, PSG6::outA1H },	// C[W
	{ 0xab, IOBus::portout, PSG6::outA3H },	// C[W
	{ 0xac, IOBus::portout, PSG6::outA0H },	// C[W
	{ 0xad, IOBus::portout, PSG6::outA1H },	// C[W
	{ 0xaf, IOBus::portout, PSG6::outA3H },	// C[W
	{ 0xa2, IOBus::portin,  PSG6::inA2H },
	{ 0xa6, IOBus::portin,  PSG6::inA2H },	// C[W
	{ 0xaa, IOBus::portin,  PSG6::inA2H },	// C[W
	{ 0xae, IOBus::portin,  PSG6::inA2H },	// C[W
	{ 0, 0, 0 }	};

// 8255(SUB CPU) -----
const IOBus::Connector VM60::c_8255s[] = {
	{ IO8049_BUS, IOBus::portout, PIO6::outPBH },
	{ IO8049_BUS, IOBus::portin,  PIO6::inPBH  },
	{ IO8049_T0,  IOBus::portin,  PIO6::inIBF  },
	{ IO8049_INT, IOBus::portin,  PIO6::inOBF  },
	{ 0, 0, 0 }	};

// 8255(Z80) -----
const IOBus::Connector VM60::c_8255m[] = {
	{ 0x90, IOBus::portout, PIO6::out90H },
	{ 0x91, IOBus::portout, PIO6::out91H },
	{ 0x92, IOBus::portout, PIO6::out92H },
	{ 0x93, IOBus::portout, PIO6::out93H },
	{ 0x94, IOBus::portout, PIO6::out90H },	// C[W
	{ 0x95, IOBus::portout, PIO6::out91H },	// C[W
	{ 0x96, IOBus::portout, PIO6::out92H },	// C[W
	{ 0x97, IOBus::portout, PIO6::out93H },	// C[W
	{ 0x98, IOBus::portout, PIO6::out90H },	// C[W
	{ 0x99, IOBus::portout, PIO6::out91H },	// C[W
	{ 0x9a, IOBus::portout, PIO6::out92H },	// C[W
	{ 0x9b, IOBus::portout, PIO6::out93H },	// C[W
	{ 0x9c, IOBus::portout, PIO6::out90H },	// C[W
	{ 0x9d, IOBus::portout, PIO6::out91H },	// C[W
	{ 0x9e, IOBus::portout, PIO6::out92H },	// C[W
	{ 0x9f, IOBus::portout, PIO6::out93H },	// C[W
	{ 0x90, IOBus::portin,  PIO6::in90H },
	{ 0x92, IOBus::portin,  PIO6::in92H },
	{ 0x93, IOBus::portin,  PIO6::in93H },
	{ 0x94, IOBus::portin,  PIO6::in90H },	// C[W
	{ 0x96, IOBus::portin,  PIO6::in92H },	// C[W
	{ 0x97, IOBus::portin,  PIO6::in93H },	// C[W
	{ 0x98, IOBus::portin,  PIO6::in90H },	// C[W
	{ 0x9a, IOBus::portin,  PIO6::in92H },	// C[W
	{ 0x9b, IOBus::portin,  PIO6::in93H },	// C[W
	{ 0x9c, IOBus::portin,  PIO6::in90H },	// C[W
	{ 0x9e, IOBus::portin,  PIO6::in92H },	// C[W
	{ 0x9f, IOBus::portin,  PIO6::in93H },	// C[W
	{ 0, 0, 0 }	};

// DISK -----
const IOBus::Connector VM60::c_disk[] = {
	{ 0xd1, IOBus::portout, DSK60::outD1H },
	{ 0xd2, IOBus::portout, DSK60::outD2H },
	{ 0xd3, IOBus::portout, DSK60::outD3H },
	{ 0xd5, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xd6, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xd7, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xd9, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xda, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xdb, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xdd, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xde, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xdf, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xd0, IOBus::portin,  DSK60::inD0H },
	{ 0xd1, IOBus::portin,  DSK60::inD1H },
	{ 0xd2, IOBus::portin,  DSK60::inD2H },
	{ 0xd3, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xd4, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xd5, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xd6, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xd7, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xd8, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xd9, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xda, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xdb, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xdc, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xdd, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xde, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xdf, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0, 0, 0 }	};

// CMT(LOAD) -----
const IOBus::Connector VM60::c_cmtl[] = {
	{ 0xb0, IOBus::portout, CMTL::outB0H },
	{ 0xb1, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb2, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb3, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb4, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb5, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb6, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb7, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb8, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xb9, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xba, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xbb, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xbc, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xbd, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xbe, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0xbf, IOBus::portout, CMTL::outB0H },	// C[W
	{ 0, 0, 0 }	};




// PC-6001mk2 --------------------------------------------------

//  -----
const IOBus::Connector VM62::c_intr[] = {
	{ 0xb0, IOBus::portout, INT62::outB0H },
	{ 0xf3, IOBus::portout, INT62::outF3H },
	{ 0xf4, IOBus::portout, INT62::outF4H },
	{ 0xf5, IOBus::portout, INT62::outF5H },
	{ 0xf6, IOBus::portout, INT62::outF6H },
	{ 0xf7, IOBus::portout, INT62::outF7H },
	{ 0xf3, IOBus::portin,  INT62::inF3H },
	{ 0xf4, IOBus::portin,  INT62::inF4H },
	{ 0xf5, IOBus::portin,  INT62::inF5H },
	{ 0xf6, IOBus::portin,  INT62::inF6H },
	{ 0xf7, IOBus::portin,  INT62::inF7H },
	{ 0, 0, 0 }	};

//  -----
const IOBus::Connector VM62::c_memory[] = {
	{ 0xc1, IOBus::portout, MEM62::outC1H },
	{ 0xc2, IOBus::portout, MEM62::outC2H },
	{ 0xc3, IOBus::portout, MEM62::outC3H },
	{ 0xf0, IOBus::portout, MEM62::outF0H },
	{ 0xf1, IOBus::portout, MEM62::outF1H },
	{ 0xf2, IOBus::portout, MEM62::outF2H },
	{ 0xf3, IOBus::portout, MEM62::outF3H },
	{ 0xf8, IOBus::portout, MEM62::outF8H },
	{ 0xf0, IOBus::portin,  MEM62::inF0H },
	{ 0xf1, IOBus::portin,  MEM62::inF1H },
	{ 0xf2, IOBus::portin,  MEM62::inF2H },
	{ 0xf3, IOBus::portin,  MEM62::inF3H },
	{ 0, 0, 0 }	};

// VDG -----
const IOBus::Connector VM62::c_vdg[] = {
	{ 0xb0, IOBus::portout, VDG62::outB0H },
	{ 0xc0, IOBus::portout, VDG62::outC0H },
	{ 0xc1, IOBus::portout, VDG62::outC1H },
	{ 0xc1, IOBus::portin,  VDG62::inC1H },
	{ 0, 0, 0 } };

// PSG -----
const IOBus::Connector VM62::c_psg[] = {
	{ 0xa0, IOBus::portout, PSG6::outA0H },
	{ 0xa1, IOBus::portout, PSG6::outA1H },
	{ 0xa3, IOBus::portout, PSG6::outA3H },
	{ 0xa2, IOBus::portin,  PSG6::inA2H },
	{ 0, 0, 0 }	};

// 8255(Z80) -----
const IOBus::Connector VM62::c_8255m[] = {
	{ 0x90, IOBus::portout, PIO6::out90H },
	{ 0x91, IOBus::portout, PIO6::out91H },
	{ 0x92, IOBus::portout, PIO6::out92H },
	{ 0x93, IOBus::portout, PIO6::out93H },
	{ 0x90, IOBus::portin,  PIO6::in90H },
	{ 0x92, IOBus::portin,  PIO6::in92H },
	{ 0x93, IOBus::portin,  PIO6::in93H },
	{ 0, 0, 0 }	};

// 8255(SUB CPU) -----
const IOBus::Connector VM62::c_8255s[] = {
	{ IO8049_BUS, IOBus::portout, PIO6::outPBH },
	{ IO8049_BUS, IOBus::portin,  PIO6::inPBH  },
	{ IO8049_T0,  IOBus::portin,  PIO6::inIBF  },
	{ IO8049_INT, IOBus::portin,  PIO6::inOBF  },
	{ 0, 0, 0 }	};

//  -----
const IOBus::Connector VM62::c_voice[] = {
	{ 0xe0, IOBus::portout, VCE6::outE0H },
	{ 0xe2, IOBus::portout, VCE6::outE2H },
	{ 0xe3, IOBus::portout, VCE6::outE3H },
	{ 0xe4, IOBus::portout, VCE6::outE0H },	// C[W
	{ 0xe6, IOBus::portout, VCE6::outE2H },	// C[W
	{ 0xe7, IOBus::portout, VCE6::outE3H },	// C[W
	{ 0xe8, IOBus::portout, VCE6::outE0H },	// C[W
	{ 0xea, IOBus::portout, VCE6::outE2H },	// C[W
	{ 0xeb, IOBus::portout, VCE6::outE3H },	// C[W
	{ 0xec, IOBus::portout, VCE6::outE0H },	// C[W
	{ 0xee, IOBus::portout, VCE6::outE2H },	// C[W
	{ 0xef, IOBus::portout, VCE6::outE3H },	// C[W
	{ 0xe0, IOBus::portin,  VCE6::inE0H },
	{ 0xe2, IOBus::portin,  VCE6::inE2H },
	{ 0xe3, IOBus::portin,  VCE6::inE3H },
	{ 0xe4, IOBus::portin,  VCE6::inE0H },	// C[W
	{ 0xe6, IOBus::portin,  VCE6::inE2H },	// C[W
	{ 0xe7, IOBus::portin,  VCE6::inE3H },	// C[W
	{ 0xe8, IOBus::portin,  VCE6::inE0H },	// C[W
	{ 0xea, IOBus::portin,  VCE6::inE2H },	// C[W
	{ 0xeb, IOBus::portin,  VCE6::inE3H },	// C[W
	{ 0xec, IOBus::portin,  VCE6::inE0H },	// C[W
	{ 0xee, IOBus::portin,  VCE6::inE2H },	// C[W
	{ 0xef, IOBus::portin,  VCE6::inE3H },	// C[W
	{ 0, 0, 0 }	};

// DISK -----
const IOBus::Connector VM62::c_disk[] = {
	{ 0xd1, IOBus::portout, DSK60::outD1H },
	{ 0xd2, IOBus::portout, DSK60::outD2H },
	{ 0xd3, IOBus::portout, DSK60::outD3H },
	{ 0xd5, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xd6, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xd7, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xd9, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xda, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xdb, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xdd, IOBus::portout, DSK60::outD1H },	// C[W
	{ 0xde, IOBus::portout, DSK60::outD2H },	// C[W
	{ 0xdf, IOBus::portout, DSK60::outD3H },	// C[W
	{ 0xd0, IOBus::portin,  DSK60::inD0H },
	{ 0xd1, IOBus::portin,  DSK60::inD1H },
	{ 0xd2, IOBus::portin,  DSK60::inD2H },
	{ 0xd3, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xd4, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xd5, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xd6, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xd7, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xd8, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xd9, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xda, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xdb, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0xdc, IOBus::portin,  DSK60::inD0H },		// C[W
	{ 0xdd, IOBus::portin,  DSK60::inD1H },		// C[W
	{ 0xde, IOBus::portin,  DSK60::inD2H },		// C[W
	{ 0xdf, IOBus::portin,  DSK60::inD2H },		// C[W?
	{ 0, 0, 0 }	};

// CMT(LOAD) -----
const IOBus::Connector VM62::c_cmtl[] = {
	{ 0xb0, IOBus::portout, CMTL::outB0H },
	{ 0, 0, 0 }	};



// PC-6601 -----------------------------------------------------

// DISK -----
const IOBus::Connector VM66::c_disk[] = {
	{ 0xb1, IOBus::portout, DSK66::outB1H },
	{ 0xb3, IOBus::portout, DSK66::outB3H },
	{ 0xd0, IOBus::portout, DSK66::outD0H },
	{ 0xd1, IOBus::portout, DSK66::outD1H },
	{ 0xd2, IOBus::portout, DSK66::outD2H },
	{ 0xd3, IOBus::portout, DSK66::outD3H },
	{ 0xd6, IOBus::portout, DSK66::outD6H },
	{ 0xd8, IOBus::portout, DSK66::outD8H },
	{ 0xda, IOBus::portout, DSK66::outDAH },
	{ 0xdd, IOBus::portout, DSK66::outDDH },
	{ 0xde, IOBus::portout, DSK66::outDEH },
	{ 0xb2, IOBus::portin,  DSK66::inB2H },
	{ 0xd0, IOBus::portin,  DSK66::inD0H },
	{ 0xd1, IOBus::portin,  DSK66::inD1H },
	{ 0xd2, IOBus::portin,  DSK66::inD2H },
	{ 0xd3, IOBus::portin,  DSK66::inD3H },
	{ 0xd4, IOBus::portin,  DSK66::inD4H },
	{ 0xdc, IOBus::portin,  DSK66::inDCH },
	{ 0xdd, IOBus::portin,  DSK66::inDDH },
	{ 0, 0, 0 }	};


