#ifndef P6VM_H_INCLUDED
#define P6VM_H_INCLUDED

#include "typedef.h"
#include "io.h"
#include "schedule.h"

#include "device/z80.h"

#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
#include "breakpoint.h"
#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //

class EL6;
class CFG6;

class CPU6;
class SUB6;
class IO6;
class MEM6;
class INT6;
class PIO6;
class VDG6;
class KEY6;
class PSG6;
class VCE6;
class CMTL;
class CMTS;
class DSK6;

class EVSC;
class BPoint;


// {z}VNX
class VM6 {
	
	friend class EL6;
	friend class DSP6;
	
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	friend class cWndMon;
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
protected:
	// foCXRlN^e[u\
	struct DEVCONNTABLE {
		const IOBus::Connector *Intr;		// 
		const IOBus::Connector *Memory;		// 
		const IOBus::Connector *Vdg;		// VDG
		const IOBus::Connector *Psg;		// PSG
		const IOBus::Connector *M8255;		// I/O(Z80)
		const IOBus::Connector *S8255;		// I/O(SUB CPU)
		const IOBus::Connector *Voice;		// 
		const IOBus::Connector *Disk;		// DISK
		const IOBus::Connector *CmtL;		// CMT(LOAD)
		const IOBus::Connector *Soldier;	// m̃J[gbW
		
		DEVCONNTABLE() : Intr(NULL), Memory(NULL), Vdg(NULL), Psg(NULL), M8255(NULL), S8255(NULL),
						 Voice(NULL), Disk(NULL), CmtL(NULL), Soldier(NULL) {}
	};
	
	
	int clock;					// CPUNbN
	
	// IuWFNg|C^
	EL6 *el;					// G~[^C
	
	EVSC *evsc;					// CxgXPW[
	IO6 *iom;					// I/O(Z80)
	IO6 *ios;					// I/O(SUB CPU)
	INT6 *intr;					// 
	CPU6 *cpum;					// CPU
	SUB6 *cpus;					// SUB CPU
	MEM6 *mem;					// 
	VDG6 *vdg;					// VDG
	PSG6 *psg;					// PSG
	VCE6 *voice;				// 
	PIO6 *pio;					// 8255
	KEY6 *key;					// L[
	CMTL *cmtl;					// CMT(LOAD)
	CMTS *cmts;					// CMT(SAVE)
	DSK6 *disk;					// DISK
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	BPoint *bp;					// u[N|Cg
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	
	DEVCONNTABLE DevTable;		// foCXRlN^e[u
	
	virtual bool AllocObjSpecific() = 0;		// @ʃIuWFNgm
	
public:
	VM6( EL6 * );								// RXgN^
	virtual ~VM6();								// fXgN^
	
	// foCXRlN^
	const static IOBus::Connector c_soldier[];	// m̃J[gbW
	
	bool AllocObject( CFG6 * );					// SIuWFNgm
	bool Init( CFG6 * );						// 
	void Reset();								// Zbg
	int Emu();									// 1ߎs
	int GetCPUClock();							// CPUNbN擾
	
	
	// P6foCXp֐Q
	// EL
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	bool ElIsMonitor();							// j^[h?
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// EVSC
	bool EventAdd( Device *, int, double, int );	// Cxgǉ
	bool EventDel( Device *, int );				// Cxg폜
	void EventUpdate( int );					// CxgXV
	void EventReset( Device::ID, int, double=0 );	// wCxgZbg
	double EventScale( Device::ID, int );		// Cxg̐is߂
	bool EventGetInfo( EVSC::evinfo * );		// Cxg擾
	bool EventSetInfo( EVSC::evinfo * );		// Cxgݒ
	void EventOnVSYNC();						// VSYNCʒm
	// IO6
	BYTE IomIn( int, int * = NULL );			// IN֐
	BYTE IosIn( int, int * = NULL );			// IN֐
	void IomOut( int, BYTE, int * = NULL );		// OUT֐
	void IosOut( int, BYTE, int * = NULL );		// OUT֐
	// INT6
	int	IntIntrCheck();							// ݃`FbN
	void IntReqIntr( DWORD );					// ݗv
	void IntCancelIntr( DWORD );				// ݓP
	// CPU6
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	int CpumDisasm( char *, WORD );				// 1CtAZu
	void CpumGetRegister( cZ80::Register * );	// WX^l擾
	void CpumSetRegister( cZ80::Register * );	// WX^lݒ
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// SUB6
	void CpusReqKeyIntr( int, BYTE );			// L[ݗv
	void CpusExtIntr();							// Oݗv
	void CpusReqCmtIntr( BYTE );				// CMT READݗv
	bool CpusIsCmtIntrReady();					// CMTݔ?
	// MEM6
	BYTE MemFetch( WORD, int * = NULL );		// tFb`(M1)
	BYTE MemRead( WORD, int * = NULL );			// [h
	void MemWrite( WORD, BYTE, int * = NULL );	// Cg
	void MemSetCGBank( bool );					// CG ROM BANK ؂ւ
	BYTE MemReadMainRom( WORD );
	BYTE MemReadMainRam( WORD );
	BYTE MemReadExtRom ( WORD );
	BYTE MemReadCGrom1 ( WORD );
	BYTE MemReadCGrom2 ( WORD );
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	const char *MemGetReadMemBlk( int );
	const char *MemGetWriteMemBlk( int );
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// VDG6
	void VdgSetCrtDisp( bool );					// CRT\Ԑݒ
	bool VdgGetWinSize();						// EBhETCY擾
	bool VdgIsDisp();							// \ԃtO擾
	// PIO6
	BYTE PioReadA();							// PartA [h
	// KEY6
	BYTE KeyGetKeyJoy();						// J[\L[Ԏ擾
	BYTE KeyGetJoy( int );						// WCXeBbNԎ擾
	BYTE KeyGetKeyIndicator();					// L[{[hCWP[^Ԏ擾
	void KeyChangeKana();						// p<->Ȑ؊
	void KeyChangeKKana();						// <->Ji؊
	// CMTL
	bool CmtlIsMount();							// }Egς?
	bool CmtlIsAutoStart();						// I[gX^[g?
	bool CmtlIsBoostUp();						// BoostUpԎ擾
	const char *CmtlGetFile();					// t@CpX擾
	const char *CmtlGetName();					// TAPE擾
	DWORD CmtlGetSize();						// x^C[WTCY擾
	int CmtlGetCount();							// JE^擾
	// CMTS
	bool CmtsMount();							// TAPE }Eg
	void CmtsUnmount();							// TAPE A}Eg
	void CmtsSetBaud( int );					// {[[gݒ
	void CmtsCmtWrite( BYTE );					// CMT 1
	// DSK6
	bool DskIsMount( int );						// }Egς?
	bool DskIsSystem( int );					// VXefBXN?
	bool DskIsProtect( int );					// veNg?
	bool DskInAccess( int );					// ANZX?
	const char *DskGetFile( int );				// t@CpX擾
	const char *DskGetName( int );				// DISK擾
	#ifndef NOMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// BPoint
	bool BpExistBreakPoint();					// u[N|Cg̗L`FbN
	bool BpCheckBreakPoint( BPoint::BPtype, WORD, BYTE, const char * );	// u[N|Cg`FbN
	void BpClearStatus();						// u[N|Cg̏NA
	BPoint::BPtype BpGetType( int );			// u[N|Cg̃^Cv擾
	void BpSetType( int, BPoint::BPtype );		// u[N|Cg̃^Cvݒ
	WORD BpGetAddr( int );						// u[N|Cg̃AhX擾
	void BpSetAddr( int, WORD );				// u[N|Cg̃AhXݒ
	bool BpIsReqBreak();						// u[Nv?
	int BpGetReqBPNum();						// u[Nv̂u[N|CgNo.擾
	void BpResetBreak();						// u[NvLZ
	#endif				// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
};


// PC-6001 z}VNX
class VM60 : public VM6 {
private:
	bool AllocObjSpecific();					// @ʃIuWFNgm
	
public:
	VM60( EL6 * );								// RXgN^
	~VM60();									// fXgN^
	
	// foCXRlN^
	const static IOBus::Connector c_intr[];		// 
	const static IOBus::Connector c_vdg[];		// VDG
	const static IOBus::Connector c_psg[];		// PSG
	const static IOBus::Connector c_8255m[];	// I/O(Z80)
	const static IOBus::Connector c_8255s[];	// I/O(SUB CPU)
	const static IOBus::Connector c_disk[];		// DISK
	const static IOBus::Connector c_cmtl[];		// CMT(LOAD)
};


// PC-6001A z}VNX
class VM61 : public VM6 {
private:
	bool AllocObjSpecific();					// @ʃIuWFNgm
	
public:
	VM61( EL6 * );								// RXgN^
	~VM61();									// fXgN^
};


// PC-6001mk2 z}VNX
class VM62 : public VM6 {
private:
	bool AllocObjSpecific();					// @ʃIuWFNgm
	
public:
	VM62( EL6 * );								// RXgN^
	~VM62();									// fXgN^
	
	// foCXRlN^
	const static IOBus::Connector c_intr[];		// 
	const static IOBus::Connector c_memory[];	// 
	const static IOBus::Connector c_vdg[];		// VDG
	const static IOBus::Connector c_psg[];		// PSG
	const static IOBus::Connector c_8255m[];	// I/O(Z80)
	const static IOBus::Connector c_8255s[];	// I/O(SUB CPU)
	const static IOBus::Connector c_voice[];	// 
	const static IOBus::Connector c_disk[];		// DISK
	const static IOBus::Connector c_cmtl[];		// CMT(LOAD)
};


// PC-6601 z}VNX
class VM66 : public VM6 {
private:
	bool AllocObjSpecific();					// @ʃIuWFNgm
	
public:
	VM66( EL6 * );								// RXgN^
	~VM66();									// fXgN^
	
	// foCXRlN^
	const static IOBus::Connector c_disk[];		// DISK
};


#endif		// P6VM_H_INCLUDED
