#include "pc6001v.h"

#include "config.h"
#include "log.h"
#include "osd.h"
#include "schedule.h"


#define WRUPDATE	(1000/SPDCNT)	/* xXVԊu(ms) */


////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
EVSC::EVSC( int mclock ) : VSYNC(false), MasterClock(mclock), NextEvent(-1), SaveClock(0) {}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
EVSC::~EVSC( void ){}


////////////////////////////////////////////////////////////////
// Cxg
//
// :	devid		foCXIuWFNgID
//			id			CxgID
// Ԓl:	evinfo *	Cxg|C^(݂ȂNULL)
////////////////////////////////////////////////////////////////
EVSC::evinfo *EVSC::Find( Device::ID devid, int id )
{
	for( int i=0; i<MAXEVENT; i++ )
		if( ev[i].devid == devid && ev[i].id == id ) return &ev[i];
	
	return NULL;
}


////////////////////////////////////////////////////////////////
// ڑfoCXo^
//
// :	dev		foCXIuWFNg|C^
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::Entry( Device *dev )
{
	// ǉ&폜ăfoCXXgɌo^
	if( !dev || !Add( dev, 0, 0, 0 ) || !Del( dev, 0 ) ) return false;
	return true;
}


////////////////////////////////////////////////////////////////
// Cxgǉ
//
// :	dev		foCXIuWFNg|C^
//			id		CxgID
//			hz		Cxgg
//			loop	[vw true:[v false:^C
//			flag	CxgX^Cw
//				bit0  : JԂw 0:^C 1:[v
//				bit2,3,4: wP 000:Hz 001:us 010:ms 100:CPUXe[g
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::Add( Device *dev, int id, double hz, int flag )
{
	// o^ς݂̏ꍇ͈U폜čēo^
	evinfo *e = Find( dev->GetID(), id );
	if( e ) Del( (Device *)devlist.Find( e->devid ), e->id );
	
	for( int i=0; i<MAXEVENT; i++ ){
		if( !ev[i].devid ){
			devlist.Add( dev );
			
			ev[i].devid  = dev->GetID();
			ev[i].id     = id;
			ev[i].Active = true;
			
			// Cxgg̐ݒ
			switch( flag&(EV_US|EV_MS|EV_STATE) ){
			case EV_US:		// usŎw
				ev[i].nps    = (double)1000000 / hz;
				ev[i].Clock  = (int)((double)MasterClock / ev[i].nps);
				break;
			case EV_MS:		// msŎw
				ev[i].nps    = (double)1000 / hz;
				ev[i].Clock  = (int)((double)MasterClock / ev[i].nps);
				break;
			case EV_STATE:	// CPUXe[gŎw
				ev[i].nps    = (double)MasterClock / hz;
				ev[i].Clock  = (int)hz;
				break;
			default:		// HzŎw
				ev[i].nps    = hz;
				ev[i].Clock  = (int)((double)MasterClock / hz);
			}
			
			// ̐ݒ
			if( flag&EV_LOOP ){	// [vCxg
				ev[i].Period = ev[i].Clock;
				if( ev[i].Period < 1 ) ev[i].Period = 1;
			}else				// ^CCxg
				ev[i].Period = 0;
			
			// ̃Cxg܂ł̃NbNXV
			if( NextEvent < 0 ) NextEvent = ev[i].Clock;
			else                NextEvent = min( NextEvent, ev[i].Clock );
			
			return true;
		}
	}
	return false;
}


////////////////////////////////////////////////////////////////
// Cxg폜
//
// :	dev		foCXIuWFNg|C^
//			id		CxgID
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::Del( Device *dev, int id )
{
	evinfo *e = Find( dev->GetID(), id );
	if( e ){
		devlist.Del( dev );
		
		e->devid  = 0;
		e->id     = 0;
		e->Active = false;
		return true;
	}else
		return false;
}


////////////////////////////////////////////////////////////////
// Cxggݒ
//
// :	devid	foCXIuWFNgID
//			id		CxgID
//			hz		Cxgg
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::SetHz( Device::ID devid, int id, double hz )
{
	evinfo *e = Find( devid, id );
	if( e ){
		e->nps    = hz;
		e->Period = (int)((double)MasterClock / hz);
		return true;
	}else
		return false;
}


////////////////////////////////////////////////////////////////
// CxgXV
//
// :	clk		i߂NbN
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void EVSC::Update( int clk )
{
	PRINTD( TIC_LOG, "[SCHE][Update] %d clock\n", clk );
	
	// NbN𗭂ߍ
	SaveClock  += clk;	// ̃Cxgp
	
	// ̃CxgNbNɒBĂȂ߂
	//  clk=0 ̏ꍇ͍XVs
	if( NextEvent > SaveClock && clk ) return;
	NextEvent = -1;
	
	int cnt;
	do{
		cnt = 0;
		for( int i=0; i<MAXEVENT; i++ ){
			// LȃCxg?
			if( ev[i].Active ){
				ev[i].Clock -= SaveClock;
				// XVԊuꍇ͕񔭐\
				// Ƃ肠SĂȂ܂ŌJԂĂƂł̂?
				if( ev[i].Clock <= 0 ){
					// CxgR[obNs
					devlist.Find( ev[i].devid )->EventCallback( ev[i].id, ev[i].Clock );
					
					if( ev[i].Period > 0 ){	// [vCxg
						ev[i].Clock += ev[i].Period;
						if( ev[i].Clock <= 0 ) cnt++;	// ̃CxgĂJEg
					}else{					// ^CCxg
						Del( (Device *)devlist.Find( ev[i].devid ), ev[i].id );
						break;
					}
				}
				// ̃Cxg܂ł̃NbNXV
				if( NextEvent < 0 ) NextEvent = ev[i].Clock;
				else                NextEvent = min( NextEvent, ev[i].Clock );
			}
		}
		SaveClock = 0;
	}while( cnt > 0 );
}


////////////////////////////////////////////////////////////////
// wCxgZbg
//
// :	devid	foCXIuWFNgID
//			id		CxgID
//			ini		JE^l(isŎw 1.0=100%)
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void EVSC::Reset( Device::ID devid, int id, double ini )
{
	PRINTD( TIC_LOG, "[SCHE][Reset] ID:%d %3f%", id, ini );
	
	evinfo *e = Find( devid, id );
	if( e ){
		// ߍ񂾃NbNl
		e->Clock = (int)((double)e->Period * ( 1.0 - min( max( 0.0, ini ), 1.0 ) )) - SaveClock;
		
		PRINTD( TIC_LOG, " -> %d/%d clock", e->Clock, e->Period );
	}
	
	PRINTD( TIC_LOG, "\n" );
}


////////////////////////////////////////////////////////////////
// Cxg܂ł̎cNbN߂
//
// :	devid	foCXIuWFNgID
//			id		CxgID
// Ԓl:	int		cXe[g(IDȂ0)
////////////////////////////////////////////////////////////////
int EVSC::Rest( Device::ID devid, int id )
{
	evinfo *e = Find( devid, id );
	if( e ) return e->Clock;
	else    return 0;
}


////////////////////////////////////////////////////////////////
// Cxg̐is߂
//
// :	devid	foCXIuWFNgID
//			id		CxgID
// Ԓl:	double	Cxgis(1.0=100%)
////////////////////////////////////////////////////////////////
double EVSC::Scale( Device::ID devid, int id )
{
	PRINTD( TIC_LOG, "[SCHE][Scale] ID:%d", id );
	
	evinfo *e = Find( devid, id );
	// Cxg݂1̃NbNݒ肳Ă?
	if( e && e->Period > 0 ){
		PRINTD( TIC_LOG, " %d/%d SAVE:%d\n", e->Clock, e->Period, SaveClock );
		
		// ߍ񂾃NbNl
		double sc = (double)( (double)( e->Period - max( e->Clock - SaveClock, 0 ) ) / (double)e->Period );
		return min( max( 0.0, sc ), 1.0 );
	}else
		return 0;
}


////////////////////////////////////////////////////////////////
// Cxg擾
//
// :	evinfo *	Cxg\̂ւ̃|C^(devid,id ZbgĂ)
// Ԓl:	bool		true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::GetEvinfo( evinfo *info )
{
	if( !info ) return false;
	
	evinfo *e = Find( info->devid, info->id );
	if( e ){
		info->Active = e->Active;
		info->Period = e->Period;
		info->Clock  = e->Clock;
		info->nps    = e->nps;
		return true;
	}
	return false;
}


////////////////////////////////////////////////////////////////
// Cxgݒ
//
// :	evinfo *	Cxg\̂ւ̃|C^(devid,id ZbgĂ)
// Ԓl:	bool		true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::SetEvinfo( evinfo *info )
{
	if( !info ) return false;
	
	// ⓚpŒǉ
	if( !Add( (Device *)devlist.Find( info->devid ), info->id, 1, EV_HZ ) ) return false;
	
	evinfo *e = Find( info->devid, info->id );
	if( e ){
		e->Active = info->Active;
		e->Period = info->Period;
		e->Clock  = info->Clock;
		e->nps    = info->nps;
		return true;
	}
	return false;
}


////////////////////////////////////////////////////////////////
// }X^NbNݒ
//
// :	clock	}X^NbN
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void EVSC::SetMasterClock( int clock )
{
	MasterClock = clock;
	for( int i=0; i<MAXEVENT; i++ )
		if( ev[i].devid && ev[i].nps > 0 && ev[i].Period > 0 ){
			ev[i].Period = (int)((double)clock / ev[i].nps);
			if( ev[i].Period < 1 ) ev[i].Period = 1;
		}

}


////////////////////////////////////////////////////////////////
// }X^NbN擾
//
// :	Ȃ
// Ԓl:	int		}X^NbN
////////////////////////////////////////////////////////////////
int EVSC::GetMasterClock( void )
{
	return MasterClock;
}


////////////////////////////////////////////////////////////////
// VSYNCɒB?
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
bool EVSC::IsVSYNC( void )
{
	return VSYNC;
}


////////////////////////////////////////////////////////////////
// VSYNCʒm
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void EVSC::OnVSYNC( void )
{
	VSYNC = true;
}


////////////////////////////////////////////////////////////////
// VSYNCtOLZ
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void EVSC::ReVSYNC( void )
{
	VSYNC = false;
}


////////////////////////////////////////////////////////////////
// ǂłSAVE
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::DokoSave( cIni *Ini )
{
	if( !Ini ) return false;
	
	Ini->PutEntry( "SCHEDULE", NULL, "MasterClock",	"%d",	MasterClock );
	Ini->PutEntry( "SCHEDULE", NULL, "VSYNC",		"%s",	VSYNC ? "Yes" : "No" );
	Ini->PutEntry( "SCHEDULE", NULL, "NextEvent",	"%d",	NextEvent );
	Ini->PutEntry( "SCHEDULE", NULL, "SaveClock",	"%d",	SaveClock );
	
	
	// Cxg
	char stren[16];
	int i = 0;
	
	while( ev[i].devid ){
		BYTE id1,id2,id3,id4;
		DWTOB( ev[i].devid, id4, id3, id2, id1 );
		sprintf( stren, "Event%02X", i );
		Ini->PutEntry( "SCHEDULE", NULL, stren, "%c%c%c%c %d %d %d %d %lf", id1, id2, id3, id4, ev[i].id, ev[i].Active ? 1 : 0, ev[i].Period, ev[i].Clock, ev[i].nps );
		i++;
	}
	
	return true;
}


////////////////////////////////////////////////////////////////
// ǂłLOAD
//
// :	Ini		INIIuWFNg|C^
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool EVSC::DokoLoad( cIni *Ini )
{
	if( !Ini ) return false;
	
	// SẴCxgЂƂ܂ɂ
	for( int i=0; i<MAXEVENT; i++ ){
		ev[i].devid  = 0;
		ev[i].id     = 0;
		ev[i].Active = false;
		ev[i].Period = 0;
		ev[i].Clock  = 0;
		ev[i].nps    = 0;
	}
	
	Ini->GetInt(    "SCHEDULE", "MasterClock",	&MasterClock,	MasterClock );
	Ini->GetTruth(  "SCHEDULE", "VSYNC",		&VSYNC, 		VSYNC );
	Ini->GetInt(    "SCHEDULE", "NextEvent",	&NextEvent,		NextEvent );	// CxgĐݒ莞ɍČ?
	Ini->GetInt(    "SCHEDULE", "SaveClock",	&SaveClock,		SaveClock );
	
	
	// Cxg
	char stren[16];
	char strrs[64];
	
	for( int i=0; i<MAXEVENT; i++ ){
		sprintf( stren, "Event%02X", i );
		if( Ini->GetString( "SCHEDULE", stren, strrs, "" ) ){
			evinfo e;
			BYTE id1,id2,id3,id4;
			int yn;
			
			sscanf( strrs,"%c%c%c%c %d %d %d %d %lf", &id1, &id2, &id3, &id4, &e.id, &yn, &e.Period, &e.Clock, &e.nps );
			e.devid = BTODW( id1, id2, id3, id4 );
			e.Active = yn ? true : false;
			if( !SetEvinfo( &e ) ) return false;
		}else
			break;
	}
	
	return true;
}









////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
SCH6::SCH6( int mclock ) : WaitEnable(true), PauseEnable(false),
	EnableScrUpdate(0),	SpeedRatio(100), SpeedCnt1(1), SpeedCnt2(1), MasterClock(mclock)
{
	INITARRAY( WRClock, 0 );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
SCH6::~SCH6( void )
{
	Stop();
}


////////////////////////////////////////////////////////////////
// Xbh֐(I[o[Ch)
//   OSD_Delay()̐xɍÊł
//   49.7ȏAғOSł͕s邯ǁDDDł
//
// :	inst	g̃IuWFNg|C^
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::OnThread( void *inst )
{
//	SCH6 *ti;
	int Vint[VSYNC_HZ];
	int VintCnt = 0;
	DWORD now,last;
	
	EnableScrUpdate = 0;
	
//	ti = STATIC_CAST( SCH6 *, inst );	// g̃IuWFNg|C^擾
	
	// 1bԂ̃C^[oݒ
	for( int i=0; i<VSYNC_HZ; i++ ) Vint[i] = (int)( 1000 / VSYNC_HZ );
	int Vrem = 1000 - (int)( 1000 / VSYNC_HZ ) * VSYNC_HZ;
	for( int i=0; i<Vrem; i++ ) Vint[(int)(VSYNC_HZ * i / Vrem)]++;
	
	// ŏ̑҂Ԃݒ
	now  = OSD_GetTicks();
	last = now;
	DWORD NextWait = now + Vint[VintCnt++];
	
	while( !this->cThread::IsCancel() ){
		now = OSD_GetTicks();
		if( now >= NextWait ){
			NextWait += Vint[VintCnt++];
			if( VintCnt >= VSYNC_HZ ){
				VintCnt -= VSYNC_HZ;
// ^C~O̓TEhXVXbhɈړ
//				ti->WaitReset();
			}
			// ʍXVtO𗧂Ă
			EnableScrUpdate++;
		}else
			OSD_Delay( 0 );
		
		if( now - last >= WRUPDATE ){
			for( int i=SPDCNT-1; i>0; i-- )
				WRClock[i] = WRClock[i-1];
			WRClock[0] = 0;
			last       += WRUPDATE;
		}
	}
}


////////////////////////////////////////////////////////////////
// Jn
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
bool SCH6::Start( void )
{
	// Xbh
	if( !this->cThread::BeginThread( this ) ) return false;
	return true;
}


////////////////////////////////////////////////////////////////
// ~
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::Stop( void )
{
	this->cThread::Cancel();	// XbhItOĂ
	this->cThread::Waiting();	// XbhI܂ő҂
	
	INITARRAY( WRClock, 0 );
}


////////////////////////////////////////////////////////////////
// WaitLtO擾
//
// :	Ȃ
// Ԓl:	L:true :false
////////////////////////////////////////////////////////////////
bool SCH6::GetWaitEnable( void )
{
	return WaitEnable;
}


////////////////////////////////////////////////////////////////
// WaitLtOݒ
//
// :	en		WaitLtO L:true :false
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::SetWaitEnable( bool en )
{
	WaitEnable = en;
}


////////////////////////////////////////////////////////////////
// |[YLtO擾
//
// :	Ȃ
// Ԓl:	L:true :false
////////////////////////////////////////////////////////////////
bool SCH6::GetPauseEnable( void )
{
	return PauseEnable;
}


////////////////////////////////////////////////////////////////
// |[YLtOݒ
//
// :	en		|[YLtO L:true :false
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::SetPauseEnable( bool en )
{
	PauseEnable = en;
}


////////////////////////////////////////////////////////////////
// VSYNC Wait
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::VWait( void )
{
	if( !WaitEnable ) return;
	
	SpeedCnt1++;
	
	if( (SpeedRatio > 100 ) && ( (SpeedCnt1 * 100) / SpeedCnt2 < SpeedRatio ) )
		return;
	
	cSemaphore::Wait();
}


////////////////////////////////////////////////////////////////
// Wait
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::WaitReset( void )
{
	if( (SpeedCnt1 > 0xffffff) || (SpeedCnt2 > 0xffffff) ){
		SpeedCnt1 /= 2;
		SpeedCnt2 /= 2;
	}
	
	SpeedCnt2++;
	
	if( (SpeedRatio < 100 ) && ( (SpeedCnt1 * 100) / SpeedCnt2 >= SpeedRatio ) )
		return;
	
	if( !cSemaphore::Value() ) cSemaphore::Post();
	
}


////////////////////////////////////////////////////////////////
// sxݒ
//
// :	spd		+:UP -:DOWN 0:
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::SetSpeedRatio( int spd )
{
	if     ( spd > 0 && SpeedRatio < 2000 ) SpeedRatio += SpeedRatio <  200 ? 10 : 100;
	else if( spd < 0 && SpeedRatio >   10 ) SpeedRatio -= SpeedRatio <= 200 ? 10 : 100;
	else if( spd == 0 )                     SpeedRatio = 100;
	
	SpeedCnt2 = VSYNC_HZ * 1000;
	SpeedCnt1 = (SpeedCnt2 * SpeedRatio) / 100;
}


////////////////////////////////////////////////////////////////
// sx擾
//
// :	Ȃ
// Ԓl:	int		sx(%)
////////////////////////////////////////////////////////////////
int SCH6::GetSpeedRatio( void )
{
	return SpeedRatio;
}


////////////////////////////////////////////////////////////////
// sx擾
//
// :	Ȃ
// Ԓl:	int		sx(%)({100)
////////////////////////////////////////////////////////////////
int SCH6::GetRatio( void )
{
	DWORD sum = 0;
	
	for( int i=1; i<SPDCNT; i++ )
		sum += WRClock[i];
	
	return (int)(double)( sum * 100.0 / MasterClock * 1000.0 / (WRUPDATE*(SPDCNT-1)) + 0.5 );
}


////////////////////////////////////////////////////////////////
// CxgXV
//
// :	clk		i߂NbN
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void SCH6::Update( int clk )
{
	WRClock[0] += clk;	// sxvZp
}


////////////////////////////////////////////////////////////////
// ʍXV}?
//
// :	Ȃ
// Ԓl:	bool		true:XV false:XVȂ
////////////////////////////////////////////////////////////////
bool SCH6::IsScreenUpdate( void )
{
	if( EnableScrUpdate ){
		EnableScrUpdate--;
		return true;
	}
	return false;
}


////////////////////////////////////////////////////////////////
// ** ǎgȂȂ܂ **
// ʂ̕\ԂɉCPUXe[gvZ
//
// :	Ȃ
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
// PC-6001mk2SR̃}jAp (ȂSR? ܂܎茳ɂꂵȂ)

// \g
//  15.70kHz(63.70usec)
//  59.92 Hz(16.69msec)

// ^C~O (P:usec)
// \[h  obN|[` \ tg|[`  v
// 32		4.87	13.74		35.76		9.33		63.70
// 40,80	4.87	 9.27		44.69		4.86		63.69

// ^C~O (P:msec)
// \[h  obN|[` \ tg|[`  v
// 192H			0.25	 2.49		12.23		1.72		16.69
// 200H			0.25	 2.23		12.74		1.47		16.69
// 204H			0.25	 2.11		12.99		1.34		16.69

// hbgNbN,CŒu
// ^C~O
// \[h  obN|[` \ tg|[`  v
// 32		34			98			256		67			455
// 40,80	34	 		66			320		35			455

// ^C~O
// \[h  obN|[` \ tg|[`  v
// 192H			3			34			192		33			262
// 200H			3		 	30			200		29			262
// 204H			4(?)		33			204		21			262


// ܂Ƃ߂ƂȊ
// g[^   455 Clk (hbgNbN 7.16MHz? 3.58MHz̗GbW?)
// \       256 320 Clk
// \     199 135 Clk
// g[^C 262 Line
// \C     192 200 204 Line
// \C    70  62  58 Line

// [\]̂CPU~ƍlƁCғ͂Ȋ
// łSR͂ȂɎ~܂Ȃ炵
// N60   : 58.769% ( 199*192 + 455*70 ) / ( 262*455 )
// N60m  : 46.313% ( 135*200 + 455*62 ) / ( 262*455 )
// N66SR : 45.239% ( 135*204 + 455*58 ) / ( 262*455 )

// CPUNbNōl
// CPUNbN/s      3993600
// CPUNbN/VSYNC    66560 (VSYNC:60HzƂ)
// CPUNbN/1Line     1109 (66560/262)
//   ғ:~/1Line  486:623 331:778
// CPUNbN/VBlank  146368  155240

/*
#define	Ref1	(double)( (double)( 199*192 + 455*70 ) / (double)( 262*455 ) )
#define	Ref2	(double)( (double)( 135*200 + 455*62 ) / (double)( 262*455 ) )

void SCH6::CalcCpuState( void )
{
	bool disp = vm->vdg->GetCrtDisp();
	bool win  = vm->vdg->GetWinSize();
	
	// 1b̃Xe[g߂
	int CpuState;
	if( disp ) CpuState = (int)( (win ? Ref1 : Ref2) * (double)CpuClock );
	else       CpuState = CpuClock;

CpuState = CpuClock;
	
	// ̂܂܂Ƒ̂œKɒ
//	CpuState = CpuState * 86 / 100;
	
	// }X^NbNݒ
	EVSC::SetMasterClock( CpuState );
}
*/

