#include <new>
#include <limits.h>

#include "common.h"
#include "vsurface.h"


////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
VSurface::VSurface( void ) : w(0), h(0), pitch(0), pixels(NULL) {}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
VSurface::~VSurface( void )
{
	if( pixels ) delete [] (BYTE *)pixels;
}



////////////////////////////////////////////////////////////////
// (T[tFXTCYw)
//
// :	ww		
//			hh		
// Ԓl:	bool	true: false:s
////////////////////////////////////////////////////////////////
bool VSurface::InitSurface( int ww, int hh )
{
	// TCY`FbN
	if( ww <= 0 || ww > 0xffff || hh <= 0 || hh > 0xffff ) return false;
	
	// 쐬ς݂Ȃ炢J
	if( pixels ) delete [] (BYTE *)pixels;
	
	// m
	pitch = (ww*sizeof(DWORD))&0xfffffffc;
	pixels = new BYTE[pitch*hh];
	if( !pixels ){
		w = h = pitch = rect.x = rect.y = rect.w = rect.h = 0;
		return false;
	}
	
	w = ww;
	h = hh;
	rect.x = rect.y = 0;
	rect.w = w;
	rect.h = h;
	
	return true;
}


////////////////////////////////////////////////////////////////
// `̈ݒ
// :	xx,yy	W
//			ww,hh	,
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void VSurface::SetRect( int xx, int yy, int ww, int hh )
{
	VRect r1,r2;
	
	r1.x = 0;
	r1.y = 0;
	r1.w = w;
	r1.h = h;
	
	r2.x = xx;
	r2.y = xx;
	r2.w = ww > 0 ? ww : 0;
	r2.h = hh > 0 ? hh : 0;
	
	RectAdd( &rect, &r1, &r2 );
}


////////////////////////////////////////////////////////////////
// `̈擾
// :	Ȃ
// Ԓl:	VRect *	`̈ւ̃|C^
////////////////////////////////////////////////////////////////
VRect *VSurface::GetRect( void )
{
	return &rect;
}


////////////////////////////////////////////////////////////////
// 擾
// :	Ȃ
// Ԓl:	int		
////////////////////////////////////////////////////////////////
int VSurface::Width( void )
{
	return w;
}


////////////////////////////////////////////////////////////////
// 擾
// :	Ȃ
// Ԓl:	int		
////////////////////////////////////////////////////////////////
int VSurface::Height( void )
{
	return h;
}


////////////////////////////////////////////////////////////////
// 1Cbyte擾
// :	Ȃ
// Ԓl:	int		1Cbyte
////////////////////////////////////////////////////////////////
int VSurface::Pitch( void )
{
	return pitch;
}


////////////////////////////////////////////////////////////////
// sNZf[^ւ̃|C^擾
// :	Ȃ
// Ԓl:	BYTE *	sNZf[^ւ̃|C^
////////////////////////////////////////////////////////////////
void *VSurface::GetPixels( void )
{
	return pixels;
}


////////////////////////////////////////////////////////////////
// _`
//
// :	x,y				W
//			col				J[R[h
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void VSurface::PSet( int x, int y, DWORD col )
{
	if( pixels && x >= rect.x && x < rect.x + rect.w && y >= rect.y && y < rect.y + rect.h ){
		DWORD *p = (DWORD *)pixels + y * pitch/sizeof(DWORD) + x;
		*p = col;
	}
}


////////////////////////////////////////////////////////////////
// F擾
//
// :	x,y				W
// Ԓl:	BYTE			J[R[h
////////////////////////////////////////////////////////////////
DWORD VSurface::PGet( int x, int y )
{
	DWORD res = 0;
	
	if( pixels && x >= rect.x && x < rect.x + rect.w && y >= rect.y && y < rect.y + rect.h ){
		DWORD *p = (DWORD *)pixels + y * pitch/sizeof(DWORD)+ x;
		res  = *p;
	}
	return res;
}


////////////////////////////////////////////////////////////////
// `̈hԂ
// :	col		F
//			rc		hԂ͈(NULLȂl)
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void VSurface::Fill( DWORD col, VRect *rc )
{
	VRect rr;
	
	if( !pixels ) return;
	
	if( rc ){
		RectAdd( &rr, &rect, rc );
	}else{
		rr.x = rect.x;
		rr.y = rect.y;
		rr.w = rect.w;
		rr.h = rect.h;
	}
	
	if( rr.w && rr.h ){
		for( int i=0; i < rr.h; i++ ){
			DWORD *p = (DWORD *)pixels + (rr.y + i) * pitch/sizeof(DWORD) + rr.x;
			for( int j=0; j < rr.w; j++ )
				*p++ = col;
		}
	}
}


////////////////////////////////////////////////////////////////
// ]
//
// :	srect			]͈
//			dst				]T[tFXւ̃|C^
//			drect			]͈(x,ŷݗL)
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void VSurface::Blit( VRect *srect, VSurface *dst, VRect *drect )
{
	VRect src1,drc1,src2,drc2;
	
	if( !dst ) return;
	
	// ]͈͐ݒ
	src1.x = srect ? srect->x : 0;
	src1.y = srect ? srect->y : 0;
	src1.w = srect ? srect->w : w;
	src1.h = srect ? srect->h : h;
	RectAdd( &src2, &src1, &rect );
	
	// ]ݒ
	drc1.x = drect ? drect->x : 0;
	drc1.y = drect ? drect->y : 0;
	drc1.w = src2.w;
	drc1.h = src2.h;
	RectAdd( &drc2, &drc1, dst->GetRect() );
	
	if( src2.w > drc2.w ) src2.w = drc2.w;
	if( src2.h > drc2.h ) src2.h = drc2.h;
	
	if( !src2.w || !src2.h || !drc2.w || !drc2.h ) return;
	
	DWORD *psrc = (DWORD *)pixels           + src2.y * pitch/sizeof(DWORD)        + src2.x;
	DWORD *pdst = (DWORD *)dst->GetPixels() + drc2.y * dst->Pitch()/sizeof(DWORD) + drc2.x;
	
	for( int i=0; i < src2.h; i++ ){
		memcpy( pdst, psrc, src2.w * sizeof(DWORD) );
		psrc += pitch/sizeof(DWORD);
		pdst += dst->Pitch()/sizeof(DWORD);
	}
}

